#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from xlettextdatabase import XletTextDatabase
from xletcalcdatabase import XletCalcDatabase
from xletdatabase import XletMemoryDatabase
from xletpatchlookup import XletPatchLookup
from . import EXTERNALS_DIR
import os


class XletCalculator(object):
    def __init__(self, dbname=None):
        ext_dir = EXTERNALS_DIR

        self.mem_db = XletMemoryDatabase("default_memory")

        self._dbs = [self.mem_db]
        self._dbs.append(XletTextDatabase(os.path.join(ext_dir, 'core/pd_objects.db'), "core"))
        self._dbs.append(XletCalcDatabase(os.path.join(ext_dir, 'core/xletsdb_core.py'), "core"))
        self._dbs.append(XletCalcDatabase(os.path.join(ext_dir, 'ceammc/xletsdb.py'), "ceammc"))

        for paths in os.walk(ext_dir):
            for text_db in [f for f in paths[2] if f.endswith(".db")]:
                extname = os.path.basename(paths[0])
                txt_db = XletTextDatabase(os.path.join(paths[0], text_db), extname)
                self._dbs.append(txt_db)

            for calc_db in [f for f in paths[2] if f == "xletsdb.py"]:
                calc_db = XletCalcDatabase(os.path.join(paths[0], calc_db), extname)
                self._dbs.append(calc_db)

        self._dbs.append(XletPatchLookup())

    def add_db(self, path, name='user'):
        self._dbs.append(XletTextDatabase(path, name))

    def has_ext_prefix(self, obj):
        return "/" in obj.name and obj.name not in ("/", "/~")

    def inlets(self, obj):
        from obj import PdObject
        if not issubclass(obj.__class__, PdObject):
            return []

        if self.has_ext_prefix(obj):
            ext, name = obj.name.split("/")[:2]
            return self.search_in_named_ext(ext, name, obj.args)[0]

        return self.inlets_by_name(obj.name, obj.args)

    def inlets_by_name(self, name, args=None):
        for db in self._dbs:
            if db.has_object(name):
                return db.inlets(name, args)
        return []

    def outlets(self, obj):
        from obj import PdObject
        if not issubclass(obj.__class__, PdObject):
            return []

        if self.has_ext_prefix(obj):
            ext, name = obj.name.split("/")[:2]
            return self.search_in_named_ext(ext, name, obj.args)[1]

        return self.outlets_by_name(obj.name, obj.args)

    def outlets_by_name(self, name, args=None):
        for db in self._dbs:
            if db.has_object(name):
                return db.outlets(name, args)

        return []

    def search_in_named_ext(self, extname, name, args):
        for db in self._dbs:
            if db.extname != extname:
                continue

            if db.has_object(name):
                return db.inlets(name, args), db.outlets(name, args)

        return [], []



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2017 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from ui_base import UIBase
from pddoc.pd import XLET_SOUND


def create_by_name(name, args=None, **kwargs):
    return UIScope(0, 0, **kwargs)


class UIScope(UIBase):
    @staticmethod
    def from_atoms(atoms):
        return UIScope(0, 0,
                    size=atoms[1],
                    min=atoms[3],
                    max=atoms[4],
                    log=atoms[5],
                    init=atoms[6],
                    send=atoms[7],
                    receive=atoms[8],
                    label=atoms[9],
                    label_xoff=atoms[10],
                    label_yoff=atoms[11],
                    font_type=atoms[12],
                    font_size=atoms[13],
                    bg_color=atoms[14],
                    fg_color=atoms[15],
                    label_color=atoms[16],
                    init_value=atoms[17],
                    steady=atoms[18])

    def __init__(self, x, y, **kwargs):
        if '@size' not in kwargs:
            kwargs['@size'] = '150x100'

        UIBase.__init__(self, "ui.scope~", x, y, **kwargs)

    def inlets(self):
        return [XLET_SOUND]

    def outlets(self):
        return []



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from coregui import CoreGui


class GCanvas(CoreGui):
    # X [size]? [width]? [height]? [send]? [receive]? [label]?
    # [x_off]? [y_off]? [font]? [font_size]? [bg_color]? [label_color]? [?]?;\r\n
    def __init__(self, x, y, **kwargs):
        CoreGui.__init__(self, "cnv", x, y, [], **kwargs)
        self._size = int(kwargs.get("size", 15))
        self.width = int(kwargs.get("width", 100))
        self.height = int(kwargs.get("height", 60))

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 13
        return GCanvas(0, 0,
                       size=atoms[0],
                       width=atoms[1],
                       height=atoms[2],
                       send=atoms[3],
                       receive=atoms[4],
                       label=atoms[5],
                       label_xoff=atoms[6],
                       label_yoff=atoms[7],
                       font_type=atoms[8],
                       font_size=atoms[9],
                       bg_color=atoms[10],
                       label_color=atoms[11])

    def draw(self, painter):
        painter.draw_rect(self.x, self.y, self.width, self.height, fill=self.bgcolor())
        self.draw_label(painter)

    def to_atoms(self):
        return ["#X", "obj", self.x, self.y, "cnv", self._size, self.width, self.height,
                self.send, self.receive,
                self.label, self._label_xoff, self._label_yoff,
                self._font_type, self._font_size, self._bg_color.to_pd(), self._label_color.to_pd(), 0]



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


class AbstractVisitor(object):
    def visit_object(self, obj):
        raise NotImplementedError

    def visit_message(self, msg):
        raise NotImplementedError

    def visit_comment(self, comment):
        raise NotImplementedError

    def visit_canvas_begin(self, canvas):
        raise NotImplementedError

    def visit_canvas_end(self, canvas):
        raise NotImplementedError

    def visit_core_gui(self, gui):
        raise NotImplementedError

    def visit_connection(self, conn):
        raise NotImplementedError

    def skip_comment(self, comment):
        return False

    def skip_object(self, obj):
        return False

    def skip_message(self, msg):
        return False

    def skip_connection(self, conn):
        return False

    def skip_core_gui(self, conn):
        return False

    def skip_canvas(self, cnv):
        return False

    def skip_children(self, obj):
        return False



#!/usr/bin/env python
# coding=utf-8
from __future__ import print_function
#   Copyright (C) 2014 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from obj import *
from . import XLET_MESSAGE, XLET_SOUND, XLET_IGNORE


class Canvas(PdObject):
    TYPE_NONE, TYPE_WINDOW, TYPE_SUBPATCH, TYPE_GRAPH = range(0, 4)

    def __init__(self, x, y, w, h, **kwargs):
        PdObject.__init__(self, "cnv", x, y, w, h)
        self._objects = []
        self._id_counter = 0
        self._name = ""
        self._type = self.TYPE_NONE
        self._graphs = []
        self._connections = {}

        if 'name' in kwargs:
            self._name = kwargs['name']

        self._font_size = kwargs.get('font_size', 12)
        self._open_on_load = kwargs.get('open_on_load', 0)

        self._graph_on_parent = False
        self._gop = {}

    @classmethod
    def subpatch(cls, name, x=0, y=0, w=200, h=100):
        c = Canvas(x, y, w, h, open_on_load=0, name=name)
        c.type = cls.TYPE_SUBPATCH
        c._graph_on_parent = False
        return c

    @classmethod
    def graph(clscls, name, x=0, y=0, w=200, h=100):
        c = Canvas(x, y, w, h, open_on_load=0, name=name)
        c.type = cls.TYPE_GRAPH
        return c

    @property
    def open_on_load(self):
        return self._open_on_load

    @property
    def font_size(self):
        return self._font_size

    @property
    def objects(self):
        return self._objects

    @property
    def graphs(self):
        return self._graphs

    @property
    def connections(self):
        return self._connections

    @property
    def type(self):
        return self._type

    @type.setter
    def type(self, t):
        assert t in (self.TYPE_NONE, self.TYPE_WINDOW, self.TYPE_SUBPATCH, self.TYPE_GRAPH)
        self._type = t

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    def append_graph(self, obj):
        assert self != obj
        assert isinstance(obj, Canvas)
        assert obj.type == self.TYPE_GRAPH
        self.append_object(obj)

    def gen_object_id(self):
        res = self._id_counter
        self._id_counter += 1
        return res

    def append_object(self, obj):
        assert self != obj
        assert issubclass(obj.__class__, BaseObject)

        if obj in self._objects:
            logging.warning("object already on canvas: {0:s}".format(obj))
            return False

        obj.id = self.gen_object_id()
        self.objects.append(obj)
        return True

    def find_object_by_id(self, oid):
        for obj in self._objects:
            if issubclass(obj.__class__, PdObject):
                if obj.id == oid:
                    return obj

        return None

    def is_graph_on_parent(self):
        return self._graph_on_parent

    def set_graph_on_parent(self, value, **kwargs):
        self._graph_on_parent = bool(value)
        if 'width' in kwargs:
            self._gop['width'] = int(kwargs['width'])
        if 'height' in kwargs:
            self._gop['height'] = int(kwargs['height'])
        if 'xoff' in kwargs:
            self._gop['xoff'] = int(kwargs['xoff'])
        if 'yoff' in kwargs:
            self._gop['yoff'] = int(kwargs['yoff'])

        self._gop['hide_args'] = int(kwargs.get('hide_args', False))

    @staticmethod
    def make_connection_key(sid, soutl, did, dinl):
        return "%i:%i => %i:%i" % (sid, soutl, did, dinl)

    def add_connection(self, sid, soutl, did, dinl, check_xlets=True):
        assert sid != did
        assert sid >= 0 and did >= 0 and soutl >= 0 and dinl >= 0

        def str_bbox(obj, xlet):
            if obj:
                return "[%s:%d]" % (obj.name, xlet)
            else:
                return "[?:%d]" % xlet

        src_obj = self.find_object_by_id(sid)
        dest_obj = self.find_object_by_id(did)

        def do_check_xlets(xlets):
            if len(xlets) < 1:
                return True

            return XLET_IGNORE not in xlets

        if src_obj and dest_obj:
            if check_xlets:
                if do_check_xlets(src_obj.outlets()):
                    if soutl >= len(src_obj.outlets()):
                        logging.warning("[{0:s}] invalid outlet number: {1:d}, total outlets: {2:d}".
                                        format(src_obj.name, soutl, len(src_obj.outlets())))
                        return False

                if do_check_xlets(dest_obj.inlets()):
                    if dinl >= len(dest_obj.inlets()):
                        logging.warning("[{0:s}] invalid inlet number: {1:d}, total inlets: {2:d}".
                                        format(dest_obj.name, dinl, len(dest_obj.inlets())))
                        return False

            ckey = Canvas.make_connection_key(sid, soutl, did, dinl)
            self._connections[ckey] = (src_obj, soutl, dest_obj, dinl)
            return True
        else:
            logging.warning("%s => %s" % (str_bbox(src_obj, soutl), str_bbox(dest_obj, dinl)))
            logging.warning("connection not found: %s:%d => %s:%d in canvas: %s" % (sid, soutl, did, dinl, self.name))
            return False

    def connect(self, args):
        assert len(args) == 4
        src_id = int(args[0])
        src_outl = int(args[1])
        dest_id = int(args[2])
        dest_inl = int(args[3])

        self.add_connection(src_id, src_outl, dest_id, dest_inl)

    def append_subpatch(self, obj):
        assert self != obj
        assert isinstance(obj, Canvas)
        assert obj.type == self.TYPE_SUBPATCH
        self.append_object(obj)
        return True

    def __str__(self):
        if self.type == self.TYPE_WINDOW:
            name = "Canvas "
        elif self.type == self.TYPE_GRAPH:
            name = "Graph "
        elif self.type == self.TYPE_SUBPATCH:
            name = "Subpatch "
        else:
            assert False

        name += "\"{0:s}\"".format(self.name)

        res = "%-30s (%i,%i %ix%i id:%i)\n" % (name, self._x, self._y, self._width, self._height, self.id)
        for obj in self.objects:
            res += "    " + str(obj)
            res += "\n"

        return res

    def gop_rect(self):
        return self._gop['xoff'], self._gop['yoff'], self._gop['width'], self._gop['height']

    def obj_is_gop(self, obj):
        x, y, w, h = self.gop_rect()
        left = x
        right = x + w
        top = y
        bottom = y + h

        if left <= obj.left and obj.right <= right and top <= obj.top and obj.bottom <= bottom:
            return True

        return False

    def draw_gop(self, painter):
        if self.type != self.TYPE_WINDOW:
            return

        painter.draw_canvas(self)
        for obj in self._objects:
            if self.obj_is_gop(obj):
                obj.draw(painter)

    def draw(self, painter):
        if self.type == self.TYPE_NONE:
            return

        if self.type == self.TYPE_SUBPATCH:
            painter.draw_subpatch(self)
        elif self.type == self.TYPE_GRAPH:
            painter.draw_graph(self)
        elif self.type == self.TYPE_WINDOW:
            painter.draw_canvas(self)

            for obj in self._objects:
                obj.draw(painter)

            painter.draw_connections(self)

            if self.is_graph_on_parent():
                painter.draw_rect(self._gop['xoff'],
                                  self._gop['yoff'],
                                  self._gop['width'],
                                  self._gop['height'],
                                  width=1,
                                  color=(1, 0.5, 0.5))

    def inlets(self):
        res = []

        objects = sorted(self._objects, key=lambda obj: obj.x)

        for o in objects:
            if issubclass(o.__class__, PdObject):
                if o.name == "inlet":
                    res.append(XLET_MESSAGE)
                elif o.name == "inlet~":
                    res.append(XLET_SOUND)

        return res

    def outlets(self):
        res = []

        objects = sorted(self._objects, key=lambda obj: obj.x)

        for o in objects:
            if issubclass(o.__class__, PdObject):
                if o.name == "outlet":
                    res.append(XLET_MESSAGE)
                elif o.name == "outlet~":
                    res.append(XLET_SOUND)

        return res

    def traverse(self, visitor):
        assert isinstance(visitor, AbstractVisitor)

        if visitor.skip_canvas(self):
            return

        if hasattr(visitor, 'visit_canvas_begin'):
            visitor.visit_canvas_begin(self)

        if not visitor.skip_children(self):
            for o in self.objects:
                o.traverse(visitor)

        for k, conn in self._connections.items():
            if not visitor.skip_connection(conn):
                visitor.visit_connection(conn)

        if hasattr(visitor, 'visit_canvas_end'):
            visitor.visit_canvas_end(self)

    def _print_connections(self):
        for k, v in self.connections.items():
            print("[{0:s}:{1:d} ({2:d},{3:d})] => [{4:s}:{5:d} ({6:d},{7:d})]".format(v[0].name, v[1], v[0].x, v[0].y,
                                                                                      v[2].name, v[3], v[2].x, v[2].y))



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                            #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import os

from mako.template import Template

import cairopainter
from docobjectvisitor import DocObjectVisitor


class HtmlDocVisitor(DocObjectVisitor):
    def __init__(self):
        DocObjectVisitor.__init__(self)
        self._image_output_dir = HtmlDocVisitor.image_output_dir
        self._image_extension = "png"
        self._image_output_dir = "img"

        self._css_file = ""
        self._css = ""
        # template config
        tmpl_path = "{0:s}/share/html_object.tmpl".format(os.path.dirname(__file__))
        self._html_template = Template(filename=tmpl_path)

    def css_file(self):
        return self._css_file

    def set_css_file(self, filename):
        self._css_file = filename

    def css(self):
        return self._css

    def set_css(self, content):
        self._css = content

    def make_image_painter(self, w, h, fname):
        return cairopainter.CairoPainter(w, h, fname, "png",
                                         xoffset=self._canvas_padding,
                                         yoffset=self._canvas_padding)

    def render(self):
        return self._html_template.render(
            title=self._title,
            description=self._description,
            keywords=self._keywords,
            css_file=self._css_file,
            css=self._css,
            aliases=self._aliases,
            license=self._license,
            version=self._version,
            examples=self._examples,
            inlets=self._inlets,
            outlets=self._outlets,
            arguments=self._arguments,
            see_also=self._see_also,
            website=self._website,
            authors=self._authors,
            contacts=self._contacts,
            library=self._library,
            category=self._category)



#!/usr/bin/env python
# coding=utf-8
#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from lxml import etree
from pd.factory import make_by_name
from pdpage import PdPage
import logging


class CategoryParser(object):
    WINDOW_WIDTH = 715
    WINDOW_HEIGHT = 500
    HEADER_HEIGHT = 40
    OBJECT_OFFSET = 30
    DESCRIPTION_OFFSET = 175

    def __init__(self, fname):
        self._fname = fname
        self._xml = None
        self._root = None
        self._cat = None
        self._cat_name = ""
        self._lib_name = ""
        self._current_y = 0
        self._pp = None

    def process(self):
        self._xml = etree.parse(self._fname)
        self._xml.xinclude()
        self._root = self._xml.getroot()
        self._lib_name = self._root.get("name")
        self.process_xml()

    def process_xml(self):
        for c in self._root.findall("category"):
            self.process_xml_category(c)

    def process_xml_category(self, c):
        self._cat = c
        self._cat_name = c.get('name')
        self._current_y = self.HEADER_HEIGHT + 10
        self._pp = PdPage("cat", self.WINDOW_WIDTH, self.WINDOW_HEIGHT)
        self.add_menubar()
        self.add_info()
        self.process_xml_category_entries()
        self.add_header()
        self.add_footer()
        self.save()
        del self._pp
        self._pp = None

    def add_info(self):
        info = self._cat.find("category-info")
        if info is not None:
            self._pp.add_description(info.text, self.HEADER_HEIGHT + 10)

    def add_menubar(self):
        l1 = self._pp.make_link(0, self._current_y, "../index-help.pd", "index")
        l2 = self._pp.make_link(0, self._current_y,
                                "{0}-help.pd".format(self._lib_name),
                                "{0}".format(self._lib_name))
        delim = self._pp.make_txt("::", 0, self._current_y)
        menu = [l1, delim, l2]
        self._pp.place_in_row(menu, 20, 8)
        self._pp.append_list(menu)
        br = self._pp.group_brect(menu)
        self._current_y += br[1] + br[3]

    def process_xml_category_entries(self):
        for e in self._cat.findall('entry'):
            self.add_object_description(e)

    def add_object_description(self, obj):
        pdobj = make_by_name(obj.get('name'))
        pdobj.x = self.OBJECT_OFFSET
        pdobj.y = self._current_y

        ref_view = obj.get('ref_view', 'object')
        if ref_view == 'object':
            self._pp.append_object(pdobj)
        elif ref_view == 'link':
            name = format(obj.get('name'))
            help_file = '{0}-help.pd'.format(name)
            link_text = '[{0}]'.format(name)
            self._pp.add_link(link_text, help_file, self.OBJECT_OFFSET, self._current_y)

        info = self._pp.add_txt(obj.get('descr'), self.DESCRIPTION_OFFSET, self._current_y)
        self._current_y += max(info.height, pdobj.height) + 10

    def save(self):
        fname = "{0}.{1}-help.pd".format(self._lib_name, self._cat_name)
        with open(fname, "w") as f:
            f.write(self._pp.to_string())
            logging.info("{0} created".format(fname))

    def add_header(self):
        title = "{0}::{1}".format(self._lib_name, self._cat_name)
        h = self._pp.add_header(title)
        return h

    def add_footer(self):
        f = self._pp.make_footer(self._current_y)
        self._pp.append_object(f)

        info = self._pp.make_txt("library: {0}".format(self._lib_name), 20, f.top + 10)
        self._pp.append_object(info)
        return f



#!/usr/bin/env python
# coding=utf-8
import argparse

from pddoc.cairopainter import *
from pddoc.pd import PdObject, BRectCalculator
from format import detect_format

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import os


def add_xlet_db(path_list):
    for db_path in path_list:
        if not os.path.exists(db_path):
            logging.warning("xlet database file not found: '%s'. skipping...", db_path)
        else:
            PdObject.xlet_calculator.add_db(db_path)


def draw_object(obj, fname, fmt, **kwargs):
    w, h = BRectCalculator().object_brect(obj)[2:]

    pad = kwargs['pad'][0]
    w += pad * 2
    h += pad * 2

    painter = CairoPainter(int(w), int(h), fname, fmt, xoffset=pad, yoffset=pad)
    painter.draw_object(obj)


def main():
    arg_parser = argparse.ArgumentParser(description='PureData patch to image converter')
    arg_parser.add_argument('--format', '-f', metavar='format', nargs=1, choices=("png", "pdf", "svg"),
                            help='output image format')
    arg_parser.add_argument('--pad', '-p', metavar='padding', nargs=1, type=int, default=[1],
                            help="adds padding around object")
    arg_parser.add_argument('name', metavar='OBJECT_NAME', help="PureData object name, for ex.: osc~")
    arg_parser.add_argument('output', metavar='OUTNAME', nargs='?', default='',
                            help="Image output name, for ex.: float.png")
    arg_parser.add_argument('--xlet-db', '-x', metavar='PATH', action='append',
                            help='inlet/outlet database file path', default=[])

    args = vars(arg_parser.parse_args())

    add_xlet_db(args['xlet_db'])

    pdo = PdObject(args['name'])

    if not args['output']:
        if args['format']:
            output = "{0:s}.{1:s}".format(args['name'], args['format'][0].lower())
        else:
            output = args['name'] + ".png"
            args['output'] = output
    else:
        output = args['output']

    fmt = detect_format(args)
    draw_object(pdo, output, fmt, pad=args['pad'])


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8
from canvas import Canvas
import textwrap
from abstractvisitor import AbstractVisitor

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


from array import Array
import re


class PdExporter(AbstractVisitor):
    def __init__(self):
        self.result = []

    def visit_canvas_begin(self, cnv):
        assert isinstance(cnv, Canvas)

        if cnv.type == Canvas.TYPE_WINDOW:
            line = "#N canvas {0:d} {1:d} {2:d} {3:d} {4};". \
                format(cnv.x, cnv.y, cnv.width, cnv.height, cnv.font_size)
            self.result.append(line)
        elif cnv.type == Canvas.TYPE_SUBPATCH:
            line = "#N canvas {0:d} {1:d} {2:d} {3:d} {4:s} 0;". \
                format(cnv.x, cnv.y, cnv.width, cnv.height, cnv.name)
            self.result.append(line)
        elif cnv.type == Canvas.TYPE_GRAPH:
            gline = "#N canvas {0:d} {1:d} {2:d} {3:d} {4:s} 0;". \
                format(cnv.x, cnv.y, cnv.width, cnv.height, cnv.name)
            self.result.append(gline)
        else:
            pass

    def visit_graph(self, graph):
        pass

    def visit_subpatch(self, spatch):
        pass

    def visit_canvas_end(self, cnv):
        if cnv.type == Canvas.TYPE_WINDOW:
            self.result.append("")
        elif cnv.type == Canvas.TYPE_SUBPATCH:
            line = "#X restore {0:d} {1:d} pd {2:s};".format(cnv.x, cnv.y, cnv.name)
            self.result.append(line)
        pass

    def visit_connection(self, conn):
        line = "#X connect {0:d} {1:d} {2:d} {3:d};".format(conn[0].id, conn[1], conn[2].id, conn[3])
        self.result.append(line)

    def visit_comment(self, comment):
        txt = re.sub('\s+', ' ', " ".join(comment.args).strip())
        line = "#X text {0:d} {1:d} {2:s};".format(comment.x, comment.y, txt)
        for l in textwrap.wrap(line):
            self.result.append(l)

    def visit_object(self, obj):
        if isinstance(obj, Array):
            line = "#N canvas {0:d} {1:d} {2:d} {3:d} {4:s} {5:d};". \
                format(obj._cnv_x, obj._cnv_y, obj._cnv_w, obj._cnv_h, "(subpatch)", 0)
            self.result.append(line)

            line = "#X array {0:s} {1:d} float {2:d};".format(obj.name, obj.size(), obj.save_flag())
            self.result.append(line)

            line = "#A 0 " + " ".join(map(lambda x: "{0:g}".format(x), obj.data())) + ";"
            for l in textwrap.wrap(line, 70):
                self.result.append(l)

            # X coords [x_from]? [y_to]? [x_to]? [y_from]? [width]? [heigth]? [graph_on_parent]?;\r\n
            line = "#X coords {0:g} {1:g} {2:g} {3:g} {4:g} {5:d} {6:d};".format(
                obj.xrange()[0], obj.yrange()[1],
                obj.xrange()[0] + obj.size(), obj.yrange()[0],
                obj.width, obj.height, 1
            )
            self.result.append(line)

            # restore
            line = "#X restore {0:d} {1:d} graph;".format(obj.x, obj.y)
            self.result.append(line)
            return
        else:
            line = "#X obj {0:d} {1:d} {2:s}".format(obj.x, obj.y, obj.name)

        if len(obj.args):
            line += " " + " ".join(obj.args)

        line += ";"

        for l in textwrap.wrap(line, 70):
            self.result.append(l)

        # handle declare
        if obj.name == "declare":
            self.result.insert(1, "#X declare {0:s};".format(" ".join(obj.args)))

    def visit_core_gui(self, gui):
        line = " ".join(map(str, gui.to_atoms())) + ";"
        for l in textwrap.wrap(line, 70):
            self.result.append(l)

    def visit_message(self, msg):
        txt = self.escape_tokens(msg.args_to_string())
        line = "#X msg {0:d} {1:d} {2:s};".format(msg.x, msg.y, txt)
        self.result.append(line)

    def save(self, fname):
        f = open(fname, 'w')
        f.write("\n".join(self.result))
        f.close()

    @classmethod
    def escape_tokens(cls, s):
        return s.replace(',', ' \\,').replace('$', '\\$')



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #


__author__ = 'Serge Poltavski'

from ui_base import UIBase
from pddoc.cairopainter import CairoPainter


def create_by_name(name, args, **kwargs):
    return UILink(0, 0, **kwargs)


class UILink(UIBase):
    def __init__(self, x, y, **kwargs):
        UIBase.__init__(self, "ui.link", x, y, **kwargs)
        self._properties.pop('@size', None)

    def inlets(self):
        return ()

    def outlets(self):
        return ()

    def url(self):
        return self._properties['@url']

    def text(self):
        if '@title' not in self._properties:
            return '<no-title>'

        return self._properties['@title']

    def draw(self, painter):
        assert isinstance(painter, CairoPainter)
        painter.draw_text(self.x + 4, self.y + 12, self.text(), color=(0, 0, 1))

    def calc_brect(self):
        x, y, w, h = self.brect_calc().string_brect(self.text(), None)
        self.width = w
        self.height = h



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from abstractvisitor import AbstractVisitor
from brectcalculator import BRectCalculator
from obj import PdObject
from message import Message
from comment import Comment
from canvas import Canvas


class ObjectBRectVisitor(AbstractVisitor):
    brect_calc = BRectCalculator()

    def __init__(self):
        self._left = None
        self._right = None
        self._top = None
        self._bottom = None

    def add_brect(self, brect):
        assert isinstance(brect, tuple)
        x, y, w, h = brect
        if self._left:
            self._left = min(self._left, x)
        else:
            self._left = x

        if self._top:
            self._top = min(self._top, y)
        else:
            self._top = y

        if self._right:
            self._right = max(self._right, x + w)
        else:
            self._right = x + w

        if self._bottom:
            self._bottom = max(self._bottom, y + h)
        else:
            self._bottom = y + h

    def clean(self):
        self._left = None
        self._right = None
        self._top = None
        self._bottom = None

    def brect(self):
        if self._left is None \
                or self._right is None \
                or self._top is None \
                or self._bottom is None:
            return None

        return self._left, self._top, self._right - self._left, self._bottom - self._top

    def visit_comment(self, comment):
        assert isinstance(comment, Comment)
        if not comment.width or not comment.height:
            br = ObjectBRectVisitor.brect_calc.comment_brect(comment)
            comment.set_width(br[2])
            comment.set_height(br[3])

        self.add_brect(comment.brect())

    def visit_object(self, obj):
        assert isinstance(obj, PdObject)
        if not obj.width or not obj.height:
            br = ObjectBRectVisitor.brect_calc.object_brect(obj)
            obj.set_width(br[2])
            obj.set_height(br[3])

        self.add_brect(obj.brect())

    def visit_message(self, msg):
        assert isinstance(msg, Message)
        if not msg.width or not msg.height:
            br = ObjectBRectVisitor.brect_calc.message_brect(msg)
            msg.set_width(br[2])
            msg.set_height(br[3])

        self.add_brect(msg.brect())

    def skip_children(self, cnv):
        assert isinstance(cnv, Canvas)

        if cnv.type != Canvas.TYPE_WINDOW:
            return True

    def skip_connection(self, conn):
        return True

    def visit_canvas_begin(self, canvas):
        assert isinstance(canvas, Canvas)
        if canvas.type != Canvas.TYPE_SUBPATCH:
            return

        # subpatch only
        # if not canvas.width or not canvas.height:
        br = ObjectBRectVisitor.brect_calc.subpatch_brect(canvas)
        canvas.set_width(br[2])
        canvas.set_height(br[3])

        self.add_brect(canvas.brect())

    def visit_canvas_end(self, canvas):
        assert isinstance(canvas, Canvas)

    def visit_core_gui(self, gui):
        self.add_brect(gui.brect())



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

import logging
import re

from . import XLET_MESSAGE, XLET_SOUND, XLET_IGNORE
from xletdatabase import XletDatabase


class XletTextDatabase(XletDatabase):
    def __init__(self, fname, extname):
        super(XletTextDatabase, self).__init__(extname)
        self._extname = extname
        self._fname = None
        self._objects = {}
        if fname:
            self.load(fname)

    def has_object(self, objname):
        return objname in self._objects

    def load(self, fname):
        self._fname = fname
        try:
            f = open(fname, "r")
            lines = f.readlines()
            for line in lines:
                self.parse(line)

        except IOError as e:
            logging.error("Load failed: {0:s}".format(fname))
            raise e

    def ignore_inlets(self, objname):
        if not self.has_object(objname):
            return False
        else:
            return XLET_IGNORE in self._objects[objname][0]

    def ignore_outlets(self, objname):
        if not self.has_object(objname):
            return False
        else:
            return XLET_IGNORE in self._objects[objname][1]

    def inlets(self, objname, args=None):
        if not self.has_object(objname):
            return []

        return self._objects[objname][0]

    def outlets(self, objname, args=None):
        if not self.has_object(objname):
            return []

        return self._objects[objname][1]

    def parse(self, line):
        if line and line[0] == "#":  # comment
            return

        atoms = re.split("\s+", line)

        if len(atoms) < 3:
            logging.error("line skip: " + line)
            return

        def parse_xlet(data_str):
            res = []
            if data_str[0] == '-':
                return res

            for char in data_str:
                if char == "~":
                    res.append(XLET_SOUND)
                elif char == ".":
                    res.append(XLET_MESSAGE)
                elif char == '?':
                    res.append(XLET_IGNORE)
                else:
                    logging.error("unknown char in inlet definition")

            return res

        objects = atoms[0].split(",")
        inl = parse_xlet(atoms[1])
        outl = parse_xlet(atoms[2])

        for obj in objects:
            if obj in self._objects:
                logging.warning("object [{0:s}] already exists in database".format(obj))
            self._objects[obj] = (inl, outl)

    def __str__(self):
        return str(self._objects)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function
import ply.lex as lex
import ply.yacc as yacc
from ply.lex import TOKEN
import re

tokens = (
    'COMMENT_START',
    'COMMENT_END',
    'ASTERISK',
    'EOL',
    'WORD',
    'TAG'
)

t_COMMENT_END = r'(\*)+\/'
t_ASTERISK = '\*'
t_EOL = r'(\r|\n|(\r\n))'
t_WORD = r'[^ \t\r\n*]+'
t_TAG = r'@[^ \t\r\n*]+'
t_ignore = ' \t'


@TOKEN(r'\/\*(\*)+?')
def t_COMMENT_START(t):
    return t


def t_error(t):
    print("Illegal character '{0:s}'".format(t.value[0]))
    t.lexer.skip(1)


def lexer():
    return lex.lex(reflags=re.UNICODE | re.DOTALL)


doc = {}


def global_tags(name, value):
    name = name[1:]
    if name in ('version', 'license', 'brief', 'author', 'depends', 'see', 'library', 'name', 'category'):
        doc[name] = value

    if name == 'inlet':
        if not name in doc:
            doc[name] = list()

        d = dict()
        d['info'] = value

        for i in ('type', 'range', 'values'):
            k = 'xlet_' + i
            if k in doc:
                d[i] = doc[k]
                doc.pop(k, None)

        doc[name].append(d)

    if name == 'outlet':
        if not name in doc:
            doc[name] = list()

        d = dict()
        d['info'] = value

        for i in ('type', 'range', 'values'):
            k = 'xlet_' + i
            if k in doc:
                d[i] = doc[k]
                doc.pop(k, None)

        doc[name].append(d)

    if name in ('type', 'range', 'values'):
        doc['xlet_' + name] = value

    doc['prev_tag'] = name


def p_pddoc(p):
    '''pddoc : pddoc eol
             | COMMENT_START pdbody COMMENT_END'''
    # javaDoc ::=  <COMMENT_START > asteriskFound (eol(line)? ) * < COMMENT_END >
    if len(p) == 3:
        if p[1]:
            p[0] = p[1]
        else:
            p[0] = p[2]
    else:
        p[0] = p[2]


def p_pdbody(p):
    '''pdbody : pdcontent'''
    if len(p) > 1:
        p[0] = p[1]


def p_pdcontent(p):
    '''pdcontent : line
                 | line pdcontent'''
    lst = filter(None, p[1:])
    p[0] = ' '.join(lst)


# line 	::= 	( word restOfLine | asterisk asteriskFound | tag tagLine )
def p_line(p):
    '''line : rest_of_line
            | ASTERISK rest_of_line'''
    if len(p) > 2:
        p[0] = p[2]


def p_rest_of_line(p):
    '''rest_of_line : eol
                    | WORD rest_of_line
                    | ASTERISK rest_of_line
                    | TAG rest_of_line'''

    ln = filter(None, p[1:])
    # last word of line
    if len(ln) == 1:
        # empty tag
        if ln[0] == '@pddoc':
            doc['pddoc'] = True
        elif ln[0] == '@pddoc_end':
            doc['pddoc'] = False
        else:
            if not 'pddoc' in doc:
                p[0] = ln[0]

    if len(ln) == 2:
        if ln[0][0] == '@':
            global_tags(p[1], p[2])
        else:
            p[0] = ' '.join(p[1:])


def p_eol(p):
    '''eol : EOL'''


def p_error(p):
    print('Unexpected token:', p)


def parse_string(data):
    do_lex(data, None)
    parser = yacc.yacc()
    txt = parser.parse(data)
    return txt, doc


def do_lex(data, func=print):
    l = lexer()
    l.input(data)

    while True:
        tok = l.token()
        if not tok:
            break
        if func:
            func(tok)


if __name__ == '__main__':
    test_data = '''/**
 *  tanh -- hyperbolic tangent function
 *  it contains
 *
   @author Serge Poltavski, Alex,
   @library ceammc
    @license GPL
    @version 1.2-beta
    @inlet test
    @outlet test outlet
    @depends czexy
 *  @see inlet
 *  @inlet float [0-1) @type float @range -1,1 @values 1,2,3,4
 *  @inlet list - range
 *  @empty_tag
    @outlet - 1 or 0, @type symbol
    @method set - sets name
    @brief - brief description @obj[test]
    @arguments args
 *


   @pddoc
   [osc~ 440]
   |\
   [dac~]
   @pddoc_end
    ***/

    '''

    print(parse_string(test_data))



#!/usr/bin/env python
# coding=utf-8
import copy

# Copyright (C) 2014 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                            #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


class LayoutItem(object):
    def __init__(self, x=0, y=0, width=0, height=0):
        self._x = x
        self._y = y
        self._w = width
        self._h = height

    def x(self):
        return self._x

    def y(self):
        return self._y

    def width(self):
        return self._w

    def height(self):
        return self._h

    def brect(self):
        return self._x, self._y, self._w, self._h

    def __str__(self):
        return "(%i,%i,%ix%i)" % (self._x, self._y, self._w, self._h)

    def right(self):
        return self._x + self._w

    def bottom(self):
        return self._y + self._h

    def top(self):
        return self._y

    def left(self):
        return self._x

    def move_by(self, x, y):
        self._x += x
        self._y += y


class Layout(object):
    HORIZONTAL, VERTICAL = range(1, 3)

    def __init__(self, direction, distance=0):
        self._parent = None
        self._elements = []
        self._dir = direction
        self._distance = distance

    @staticmethod
    def vertical(distance=0):
        return Layout(Layout.VERTICAL, distance)

    @staticmethod
    def horizontal(distance=0):
        return Layout(Layout.HORIZONTAL, distance)

    def is_vertical(self):
        return self._dir == Layout.VERTICAL

    def is_horizontal(self):
        return self._dir == Layout.HORIZONTAL

    def parent(self):
        return self._parent

    def set_parent(self, parent):
        self._parent = parent

    def count(self):
        return len(self._elements)

    def _add_vertical(self, item):
        if self.count() > 0:
            prev_item = self._elements[-1]
            item.move_by(0, prev_item.bottom() + self._distance)

        self._elements.append(item)

    def _add_horizontal(self, item):
        if self.count() > 0:
            prev_item = self._elements[-1]
            item.move_by(prev_item.right() + self._distance, 0)

        self._elements.append(item)

    def add_item(self, item):
        assert item != self

        if item in self._elements:
            item2 = copy.copy(item)
            return self.add_item(item2)

        assert isinstance(item, LayoutItem)
        if self._dir == self.HORIZONTAL:
            self._add_horizontal(item)
        elif self._dir == self.VERTICAL:
            self._add_vertical(item)
        else:
            assert False

    def pop_item(self):
        return self._elements.pop()

    def add_layout(self, layout):
        assert layout != self

        if layout in self._elements:
            layout2 = copy.deepcopy(layout)
            return self.add_layout(layout2)

        if self.count() > 0:
            prev_item = self._elements[-1]

            if self._dir == self.HORIZONTAL:
                layout.move_by(prev_item.right() + self._distance, 0)
            elif self._dir == self.VERTICAL:
                layout.move_by(0, prev_item.bottom() + self._distance)
            else:
                assert False

        self._elements.append(layout)

    def item(self, num):
        assert num < len(self._elements)
        return self._elements[num]

    def brect(self):
        if self.count() == 0:
            return 0, 0, 0, 0

        left = []
        top = []
        right = []
        bottom = []

        for item in self._elements:
            br = item.brect()
            left.append(br[0])
            top.append(br[1])
            right.append(br[0] + br[2])
            bottom.append(br[1] + br[3])

        x0 = min(left)
        y0 = min(top)
        x1 = max(right)
        y1 = max(bottom)
        return x0, y0, x1 - x0, y1 - y0

    def x(self):
        return self.brect()[0]

    def y(self):
        return self.brect()[1]

    def width(self):
        return self.brect()[2]

    def height(self):
        return self.brect()[3]

    def top(self):
        return self.y()

    def left(self):
        return self.x()

    def bottom(self):
        br = self.brect()
        return br[1] + br[3]

    def right(self):
        br = self.brect()
        return br[0] + br[2]

    def move_by(self, x, y):
        for item in self._elements:
            item.move_by(x, y)

    def __str__(self):
        res = ""
        if self._dir == self.HORIZONTAL:
            s = ""
            for item in self._elements:
                s += "[%i:%ix%i] " % (item.x(), item.width(), item.height())

            s = s.strip()
            res += "-" * (len(s) + 4)
            res += "\n| {0:s} |\n".format(s)
            res += "-" * (len(s) + 4)
        elif self._dir == self.VERTICAL:
            col_len = 0
            els = []
            for item in self._elements:
                els.append("[%i:%ix%i]" % (item.y(), item.width(), item.height()))
                col_len = max(col_len, len(els[-1]))

            res += "-" * (col_len + 4)
            res += "\n"
            fmt = "| %{0:d}s |\n".format(col_len)
            for e in els:
                res += fmt % (e)
            res += "-" * (col_len + 4)

        return res



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import pd
from layout import *
from docobject import DocPdobject, DocPdmessage
import logging


class PdLayout(object):
    def __init__(self):
        self._canvas = None
        self._cur_layout = []
        self._example_brect = ()
        self._pdobj_id_map = {}
        self._brect_calc = pd.BRectCalculator()
        self._comment_xoffset = 2
        self._hlayout_space = 20
        self._vlayout_space = 25

    def canvas_brect(self):
        return 0, 0, self._canvas.width, self._canvas.height

    def layout_brect(self):
        return self._example_brect

    @property
    def canvas(self):
        return self._canvas

    @canvas.setter
    def canvas(self, cnv):
        assert isinstance(cnv, pd.Canvas)
        self._canvas = cnv

    def update(self):
        for pdo in self._canvas.objects:
            if not hasattr(pdo, "layout"):
                continue

            litem = getattr(pdo, "layout")
            pdo.x = litem.x()
            pdo.y = litem.y()

    def calc_brect(self, obj):
        if isinstance(obj, pd.Message):
            return self._brect_calc.message_brect(obj)
        elif isinstance(obj, pd.PdObject):
            return self._brect_calc.object_brect(obj)
        elif isinstance(obj, pd.Comment):
            return self._brect_calc.comment_brect(obj)
        else:
            assert False

    def row_begin(self):
        lh = Layout(Layout.HORIZONTAL, self._hlayout_space)
        self._cur_layout.append(lh)

    def row_end(self):
        lh = self._cur_layout.pop()
        # adds child layout to parent
        if self._cur_layout:
            self._cur_layout[-1].add_layout(lh)

        self._example_brect = lh.brect()

    def col_begin(self):
        lv = Layout(Layout.VERTICAL, self._vlayout_space)
        self._cur_layout.append(lv)

    def col_end(self):
        lv = self._cur_layout.pop()
        # adds child layout to parent
        if self._cur_layout:
            self._cur_layout[-1].add_layout(lv)

        self._example_brect = lv.brect()

    def comment2pd_comment(self, txt):
        pd_comment = pd.Comment(0, 0, txt.split(" "))
        cbbox = self.calc_brect(pd_comment)
        comment_litem = LayoutItem(0, 0, cbbox[2], cbbox[3])
        setattr(pd_comment, "layout", comment_litem)
        return pd_comment

    def doc2msg(self, doc_msg):
        assert isinstance(doc_msg, DocPdmessage)
        pdm = pd.Message(0, 0, [doc_msg.text()])
        obj_bbox = list(self.calc_brect(pdm))
        litem = LayoutItem(doc_msg.offset(), 0, obj_bbox[2], obj_bbox[3])
        setattr(pdm, "layout", litem)
        return pdm

    def doc2obj(self, doc_obj):
        assert isinstance(doc_obj, DocPdobject)
        args = filter(None, doc_obj.args())
        pd_obj = pd.factory.make_by_name(doc_obj.name(), args, **doc_obj.attrs())
        obj_bbox = list(self.calc_brect(pd_obj))
        litem = LayoutItem(doc_obj.offset(), 0, obj_bbox[2], obj_bbox[3])
        setattr(pd_obj, "layout", litem)

        if doc_obj.highlight():
            setattr(pd_obj, "highlight", True)

        return pd_obj

    def connect_begin(self, c):
        try:
            src_id = self._pdobj_id_map[c.src_id()]
            dest_id = self._pdobj_id_map[c.dest_id()]
            src_out = c._src_out
            dest_in = c._dest_in

            self._canvas.add_connection(src_id, src_out, dest_id, dest_in)
        except KeyError as e:
            logging.warning("connection not found: {0:s}:{1:s} => {2:s}:{3:s}".
                            format(c.src_id(), src_out, c.dest_id(), dest_in))

    def message_begin(self, msg_obj):
        pd_msg = self.doc2msg(msg_obj)
        self._canvas.append_object(pd_msg)

        # handle object comment
        if msg_obj.comment:
            pd_comment = self.comment2pd_comment(msg_obj.comment)
            self._canvas.append_object(pd_comment)

            hor_layout = Layout.horizontal(self._comment_xoffset)
            hor_layout.add_item(pd_msg.layout)
            hor_layout.add_item(pd_comment.layout)
            self._cur_layout[-1].add_layout(hor_layout)
        else:
            self._cur_layout[-1].add_item(pd_msg.layout)

        self.add_id_mapping(msg_obj, pd_msg)

    def object_begin(self, doc_obj):
        pd_obj = self.doc2obj(doc_obj)
        self._canvas.append_object(pd_obj)

        # handle object comment
        if doc_obj.comment:
            pd_comment = self.comment2pd_comment(doc_obj.comment)
            self._canvas.append_object(pd_comment)

            hor_layout = Layout.horizontal(self._comment_xoffset)
            hor_layout.add_item(pd_obj.layout)
            hor_layout.add_item(pd_comment.layout)
            self._cur_layout[-1].add_layout(hor_layout)
        else:
            self._cur_layout[-1].add_item(pd_obj.layout)

        self.add_id_mapping(doc_obj, pd_obj)

    def add_id_mapping(self, doc_obj, pd_obj):
        self._pdobj_id_map[doc_obj.id] = pd_obj.id

    def comment(self, doc_comment):
        comments = doc_comment.text().split("\n")

        vlayout = Layout.vertical()

        for c in comments:
            pd_c = self.comment2pd_comment(c)
            self._canvas.append_object(pd_c)
            vlayout.add_item(pd_c.layout)

        self._cur_layout[-1].add_layout(vlayout)



#!/usr/bin/env python
# coding=utf-8
from obj import PdObject

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'


class Array(PdObject):
    def __init__(self, name, size, save=0, cnv_x=0, cnv_y=22, cnv_w=450, cnv_h=300):
        PdObject.__init__(self, name)
        self.width = 200
        self.height = 140
        self._fixed_size = True
        self._size = int(float(size))
        self._save = int(save)
        self._data = [0] * self._size
        self._xrange = (0, self._size)
        self._yrange = (-1, +1)
        self._cnv_x = cnv_x
        self._cnv_y = cnv_y
        self._cnv_w = cnv_w
        self._cnv_h = cnv_h
        self.calc_brect()

    def save_flag(self):
        return self._save

    def size(self):
        return self._size

    def data(self):
        return self._data

    def set_data(self, atoms):
        assert len(atoms) == self._size
        self._data = map(lambda x: float(x), atoms)

    def xrange(self):
        return self._xrange

    def set_xrange(self, min_x, max_x):
        self._xrange = (min_x, max_x)

    def yrange(self):
        return self._yrange

    def set_yrange(self, min_y, max_y):
        self._yrange = (min_y, max_y)

    @staticmethod
    def from_atoms(atoms):
        assert len(atoms) == 4 or len(atoms) == 6
        return Array(atoms[0], atoms[1], atoms[3])

    def draw(self, painter):
        painter.draw_rect(*self.brect())
        painter.draw_text(self.x, self.y - 5, self.name)

        if not self._data:
            return
        
        swd = self.width / self._size
        yrange = self._yrange[1] - self._yrange[0]
        sht = self.height / yrange
        for idx in range(0, self._size):
            x0 = self.left + swd * idx
            x1 = x0 + swd
            y = self.top + sht * self._data[idx] + (self.height * self._yrange[1])/yrange
            painter.draw_line(x0, y, x1, y, width=3, color=(0, 0, 0))

    def to_string(self):
        res = PdObject.unescape(self.name) + ' ' + self.args_to_string()
        return res.strip()

    def inlets(self):
        return []

    def outlets(self):
        return []




#!/usr/bin/env python
# coding=utf-8
from __future__ import print_function
import argparse
import os
import logging

from pddoc.docobject import DocObject
from pddoc.pddocvisitor import PdDocVisitor
from pddoc.xletdocvisitor import XletDocVisitor
from pddoc.parser import parse_xml
from pddoc.pd.obj import PdObject
from pddoc.pd import factory

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


def add_xlet_db(path_list):
    for db_path in path_list:
        if not os.path.exists(db_path):
            logging.warning("xlet database file not found: '%s'. skipping...", db_path)
        else:
            PdObject.xlet_calculator.add_db(db_path)


def main():
    arg_parser = argparse.ArgumentParser(description='PureData pddoc to pd patch converter')
    arg_parser.add_argument('--website', '-w', metavar='URL', help='library website URL')
    arg_parser.add_argument('--license', '-l', metavar='license', help='library license')
    arg_parser.add_argument('--version', '-v', metavar='version', default='0.0', help='library version')
    arg_parser.add_argument('--force', '-f', action='store_true', help='force to overwrite existing file')
    arg_parser.add_argument('--xlet-db', metavar='PATH', action='append',
                            help='inlet/outlet database file path', default=[])
    arg_parser.add_argument('name', metavar='PDDOC', help="Documentation file in PDDOC(XML) format")
    arg_parser.add_argument('output', metavar='OUTNAME', nargs='?', default='',
                            help="Pd output patch file name")

    args = vars(arg_parser.parse_args())
    in_file = args['name']
    output = args['output']

    if not output:
        output = os.path.splitext(os.path.basename(in_file))[0] + "-help.pd"

    if os.path.exists(output) and not args['force']:
        print("Error: file already exists: '{0}'. Use --force flag to overwrite.".format(output))
        exit(1)

    add_xlet_db(args['xlet_db'])

    xml = parse_xml(in_file)

    if not xml:
        exit(1)

    factory.add_import('ceammc')

    pddoc = xml.getroot()
    for child_tag in pddoc:
        if child_tag.tag == "object":
            dobj = DocObject()
            dobj.from_xml(child_tag)

            x = XletDocVisitor()
            dobj.traverse(x)

            v = PdDocVisitor()

            if 'version' in args:
                v._version = args['version']

            if 'license' in args:
                v._license['name'] = args['license']

            if 'website' in args:
                v._website = args['website']

            dobj.traverse(v)

            patch_data = v.render()
            if not patch_data:
                print("convertion error")
                exit(1)

            with open(output, 'w') as f:
                f.write(patch_data)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import re
import os
import logging

import pddoc.pd as pd
from pdlayout import PdLayout
from idocobjectvisitor import IDocObjectVisitor
from txt.parser import Parser
from pd.canvas import Canvas
from pd.pdexporter import PdExporter
from pd.obj import PdObject


class DocObjectVisitor(IDocObjectVisitor):
    def __init__(self):
        self._title = ""
        self._description = ""
        self._keywords = []
        self._website = ""
        self._library = ""
        self._category = ""
        self._version = ""
        self._license = {}
        self._aliases = []
        self._see_also = []
        self._examples = []
        self._authors = []
        self._contacts = ""
        self._inlets = {}
        self._outlets = {}
        self._arguments = []
        self._props = []
        self._inlet_counter = 0
        self._image_counter = 0
        self._image_prefix = ""
        self._image_extension = ""
        self._image_output_dir = ""
        self._search_dir = ""
        self._layout = PdLayout()
        self._canvas_padding = 10
        self._image_object_padding = 4
        self._pdascii = ""
        self._inlet_idx = 0
        self._outlet_idx = 0
        self._since = ""
        self._is_gui = False

    def object_begin(self, obj):
        self._is_gui = obj.is_gui()

    def aliases_begin(self, a):
        if not a.aliases():
            return

        for alias in a.aliases() + [self._title]:
            self.add_alias(alias)

    def add_alias(self, name):
        element = {
            'name': name,
            'image': self.image_pdobject_fname(name)
        }
        self._aliases.append(element)

    def title_begin(self, t):
        self._title = t.text()

    def website_begin(self, w):
        self._website = w.text()

    def keywords_begin(self, k):
        self._keywords = k.keywords()

    def description_begin(self, d):
        self._description = d.text()

    def license_begin(self, l):
        self._license['url'] = l.url()
        self._license['name'] = l.name()

    def library_begin(self, lib):
        self._library = lib.text()

    def category_begin(self, cat):
        self._category = cat.text()

    def version_begin(self, v):
        self._version = v.text()

    def author_begin(self, author):
        self._authors.append(author.text())

    def contacts_begin(self, cnt):
        self._contacts = cnt.text()

    def pdexample_begin(self, tag):
        self._layout.canvas = pd.Canvas(0, 0, 10, 10, name="10")
        self._layout.canvas.type = pd.Canvas.TYPE_WINDOW

    def pdexample_end(self, tag):
        img_id, img_path = self.make_image_id_name()
        # append data to template renderer
        self._pd_append_example(img_id, img_path, None, tag.title())
        # update layout - place all objects
        self._layout.update()
        # draw image
        w, h = self._pd_layout_size(tag)
        self._pd_draw(w, h, img_path)

    def pdcomment_begin(self, comment):
        self._layout.comment(comment)

    def pdinclude_begin(self, tag):
        # assert isinstance(tag, DocPdinclude)

        pd_file_path = os.path.join(self._search_dir, tag.file())

        if not os.path.exists(pd_file_path):
            logging.error("Error in tag <pdinclude>: file not exists: \"{0:s}\"".format(pd_file_path))
            return

        parser = pd.Parser()
        if not parser.parse(pd_file_path):
            logging.error("Error in tag <pdexample>: can't process file: {0:s}".format(pd_file_path))
            return

        self._layout.canvas = parser.canvas
        img_id, img_path = self.make_image_id_name()
        # append data to template renderer
        self._pd_append_example(img_id, img_path, pd_file_path, pd_file_path)

        # TODO auto layout
        w, h = self._layout.canvas_brect()[2:]
        self._pd_draw(w, h, img_path)

    def row_begin(self, row):
        self._layout.row_begin()

    def row_end(self, row):
        self._layout.row_end()

    def col_begin(self, col):
        self._layout.col_begin()

    def col_end(self, col):
        self._layout.col_end()

    def pdascii_begin(self, pdascii):
        self._pdascii = pdascii.text()

        p = Parser()
        p.X_PAD = pdascii.x_pad()
        p.Y_PAD = pdascii.y_pad()
        p.X_SPACE *= pdascii.x_space()
        p.Y_SPACE *= pdascii.y_space()
        if not p.parse(self._pdascii):
            logging.info("<pdascii> parse failed: {0}".format(self._pdascii))

        cnv = Canvas(0, 0, 300, 500)
        cnv.type = Canvas.TYPE_WINDOW
        p.export(cnv)

        # save as pd
        br_calc = cnv.brect_calc()
        cnv.traverse(br_calc)
        bbox = br_calc.brect()
        wd = bbox[2] + Parser.X_PAD * 2
        ht = bbox[3] + Parser.Y_PAD * 2
        cnv.height = ht
        cnv.width = wd
        pd_exporter = PdExporter()
        cnv.traverse(pd_exporter)

        pd_name_template = "{0}-example.pd".format(self.name)
        pd_exporter.save(pd_name_template)

        return cnv

    def pdmessage_begin(self, msg_obj):
        self._layout.message_begin(msg_obj)

    def pdobject_begin(self, doc_obj):
        self._layout.object_begin(doc_obj)

    def pdconnect_begin(self, c):
        self._layout.connect_begin(c)

    def arguments_begin(self, args):
        self._arguments = args.items()

    def properties_begin(self, props):
        self._props = props.items()

    def since_begin(self, s):
        self._since = s.text()

    def see_begin(self, see):
        element = {
            'name': see.text(),
            'image': self.image_pdobject_fname(see.text()),
            'is_link': see.is_link()
        }

        self._see_also.append(element)

    def set_image_prefix(self, prefix):
        self._image_prefix = re.sub('[^a-zA-Z0-9~]', '', prefix)

    def image_prefix(self):
        if self._image_prefix:
            return self._image_prefix + "_"
        else:
            return ""

    def image_output_dir(self):
        return self._image_output_dir

    def set_image_output_dir(self, path):
        self._image_output_dir = path

    def search_dir(self):
        return self._search_dir

    def set_search_dir(self, path):
        self._search_dir = path

    def image_pdobject_fname(self, name):
        return os.path.join(self._image_output_dir,
                            "object_{0:s}.{1:s}".format(name, self._image_extension))

    def create_image_output_dir(self):
        try:
            if not os.path.exists(self._image_output_dir):
                os.makedirs(self._image_output_dir)

            if not os.path.isdir(self._image_output_dir):
                raise RuntimeError("not a directory: %s".format(self._image_output_dir))
        except Exception as e:
            raise RuntimeError(e.message)

    def make_image_id_name(self):
        self._image_counter += 1
        cnt = self._image_counter
        path = os.path.join(self._image_output_dir,
                            "{1:s}image_{0:02d}.{2:s}".format(self._image_counter,
                                                              self.image_prefix(),
                                                              self._image_extension))
        return cnt, path

    def make_image_painter(self, w, h, fname):
        return None

    def _pd_draw(self, w, h, fname):
        self.create_image_output_dir()

        painter = self.make_image_painter(w, h, fname)
        self._layout.canvas.draw(painter)
        logging.info("image [{0:d}x{1:d}] saved to: \"{2:s}\"".format(w, h, fname))

    def _pd_layout_size(self, tag):
        w = tag.width()
        h = tag.height()
        if not w:
            w = self._layout.layout_brect()[2]
        if not h:
            h = self._layout.layout_brect()[3]

        return int(w + 2 * self._canvas_padding), int(h + 2 * self._canvas_padding)

    def _pd_append_example(self, img_id, img_path, pd_path="", title=""):
        example_dict = {
            'id': img_id,
            'image': img_path,
            'title': title,
            'file': pd_path
        }

        self._examples.append(example_dict)

    def generate_object_image(self, name):
        fname = self.image_pdobject_fname(name)
        if os.path.exists(fname):
            logging.warning("image exists: \"%s\"", fname)

        pdo = pd.make_by_name(name)
        x, y, w, h = pd.BRectCalculator().object_brect(pdo)
        w = int(w + self._image_object_padding * 2)
        h = int(h + self._image_object_padding * 2)
        old_pad = self._canvas_padding
        self._canvas_padding = self._image_object_padding
        painter = self.make_image_painter(w, h, fname)
        self._canvas_padding = old_pad
        pdo.draw(painter)

    def generate_images(self):
        try:
            self.create_image_output_dir()

            if self._aliases:
                for a in self._aliases:
                    self.generate_object_image(a['name'])

            if self._see_also:
                for sa in self._see_also:
                    self.generate_object_image(sa['name'])
        except Exception as e:
            logging.error("Error while generating images: %s", e)



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from sys import platform as _platform

from . import XLET_MESSAGE, XLET_GUI, XLET_SOUND


class DrawStyle(object):
    fill_color = property(lambda self: (1, 1, 1))
    font_size = property(lambda self: 12)

    obj_fill_color = property(lambda self: (0.95, 0.95, 0.95))
    obj_text_color = property(lambda self: (0, 0, 0))
    obj_border_color = property(lambda self: (0.2, 0.2, 0.2))
    obj_line_width = property(lambda self: 1)
    obj_pad_x = property(lambda self: 2.5)
    obj_pad_y = property(lambda self: 2)
    obj_min_width = property(lambda self: 22)

    msg_fill_color = property(lambda self: 0.94)
    msg_min_width = property(lambda self: 22)

    xlet_width = property(lambda self: 8)
    xlet_msg_height = property(lambda self: 2)
    xlet_snd_height = property(lambda self: 2)
    xlet_gui_height = property(lambda self: 1)
    xlet_msg_color = property(lambda self: (0, 0, 0))
    xlet_snd_color = property(lambda self: (0.3, 0.2, 0.4))
    xlet_gui_color = property(lambda self: (0, 0, 0))

    conn_snd_width = property(lambda self: 2)
    conn_snd_stripe = property(lambda self: True)
    conn_snd_color = property(lambda self: (0.2, 0.2, 0.2))
    conn_snd_color2 = property(lambda self: (0.6, 0.6, 0))
    conn_snd_dash = property(lambda self: [4, 8])

    conn_msg_color = property(lambda self: (0, 0, 0))
    conn_msg_width = property(lambda self: 1)

    comment_color = property(lambda self: (0.5, 0.5, 0.5))

    highlight_color = property(lambda self: (1, 0.9, 0.8))
    highlight_padding = 3

    @property
    def obj_height(self):
        return self.font_size + 5

    def xlet_height(self, t):
        if t == XLET_MESSAGE:
            return self.xlet_msg_height
        elif t == XLET_GUI:
            return self.xlet_gui_height
        elif t == XLET_SOUND:
            return self.xlet_snd_height
        else:
            assert False

    @property
    def font_family(self):
        # FIXME
        return "Menlo"

        if _platform == "darwin":
            return "Menlo"
        else:
            return "terminus"



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'


from pddoc.pd import PdObject
from pddoc.cairopainter import CairoPainter


def create(atoms):
    assert len(atoms)
    return PddpDsp(atoms[0], atoms[1:])


class PddpDsp(PdObject):
    def __init__(self, name, args):
        PdObject.__init__(self, name, 0, 0, 0, 0, args)

    def inlets(self):
        return ()

    def outlets(self):
        return ()

    def url(self):
        return self.args[0]

    def get_height(self):
        return 18

    def get_width(self):
        return 60

    def draw(self, painter):
        assert isinstance(painter, CairoPainter)
        painter.draw_rect(self.x, self.y, self.width, self.height, color=(0, 0, 0), fill=(0.5, 1, 0.5), width=1)
        line_x = int(self.left + 18)
        painter.draw_line(line_x, self.top, line_x, self.bottom, color=(0, 0, 0), width=1)
        painter.draw_text(line_x + 5, self.bottom - 3, "dsp", color=(0, 0, 0), font_size=14, font="Monospace")


#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'


class XletDatabase(object):
    def __init__(self, extname=None):
        self._extname = extname

    def has_object(self, name):
        return False

    def ignore_inlets(self, objname):
        return False

    def ignore_outlets(self, objname):
        return False

    def inlets(self, name, args=None):
        if args is None:
            args = []
        return []

    def outlets(self, name, args=None):
        if args is None:
            args = []
        return []

    @property
    def extname(self):
        return self._extname


class XletMemoryDatabase(XletDatabase):
    def __init__(self, extname):
        super(XletMemoryDatabase, self).__init__(extname)
        self._extname = extname
        self._objects = {}

    def has_object(self, objname):
        return objname in self._objects

    def inlets(self, objname, args=None):
        if args is None:
            args = []
        if not self.has_object(objname):
            return []

        return self._objects[objname][0]

    def outlets(self, objname, args=None):
        if args is None:
            args = []
        if not self.has_object(objname):
            return []

        return self._objects[objname][1]

    def add_object(self, name, inlets, outlets):
        self._objects[name] = (inlets, outlets)

    def remove_object(self, name):
        del self._objects[name]

    def set_inlets(self, name, inlets):
        self._objects[name][0] = inlets

    def set_outlets(self, name, outlets):
        self._objects[name][1] = outlets



#!/usr/bin/env python
# coding=utf-8
from lxml import etree
import logging
from pddoc.parser import get_parser
import os
#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

import os.path


class LibraryMaker(object):
    NSMAP = {'xi': "http://www.w3.org/2001/XInclude"}

    def __init__(self, name):
        etree.register_namespace("xi", "http://www.w3.org/2001/XInclude")
        self._name = name
        self._lib = etree.Element("library", version="1.0", name=name, nsmap=self.NSMAP)
        self._cats = {}
        self._cat_entries = {}
        self._authors = {}
        self._version = ""
        self._search_paths = []

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, v):
        self._version = v

    def add_search_path(self, path):
        if not os.path.exists(path) or not os.path.isdir(path):
            logging.error("invalid search path added: {0:s}".format(path))
            return False

        if path in self._search_paths:
            logging.warning("search path already exists in search path list: {0:s}".format(path))
            return False

        self._search_paths.append(path)
        return True

    def find_in_path(self, path):
        if os.path.exists(path):
            return path

        for p in self._search_paths:
            new_path = os.path.join(p, path)
            if os.path.exists(new_path):
                return new_path

        return False

    def process_files(self, files):
        for f in files:
            self.process_object_file(f)

        self.fill_library_meta()

    def process_object_file(self, f):
        try:
            xml = etree.parse(f, get_parser())
        except etree.XMLSyntaxError as e:
            logging.error("XML syntax error:\n \"%s\"\n\twhile parsing file: \"%s\"", e, f)
            return

        pddoc = xml.getroot()
        # find all object tags, but normally it's only one per file
        objects = filter(lambda x: x.tag == 'object', pddoc)
        for obj in objects:
            self.process_object_doc(f, obj)

    def process_object_doc(self, doc_fname, xml_obj):
        name = xml_obj.get('name')
        descr = ' '.join(xml_obj.find('meta/description').text.split())
        # logging.info("[%s] doc found: %s", name, descr.text)
        categ = xml_obj.find('meta/category')
        if categ is None:
            self.add_to_others(doc_fname, name=name, descr=descr)
        else:
            self.add_to_cat(categ.text, doc_fname, name=name, descr=descr, ref_view=categ.get('view', 'object'))
        lib = xml_obj.find('meta/library').text
        if lib != self._name:
            logging.warning("library differs in file: '%s': %s != %s", doc_fname, self._name, lib)
        authors = xml_obj.findall('meta/authors/author')
        for a in authors:
            self._authors[a.text] = True

    def fill_library_meta(self):
        # process library meta
        meta = etree.Element("meta")
        version = etree.Element("version")
        version.text = self._version
        meta.append(version)
        authors = etree.Element("authors")
        for a in sorted(self._authors.keys()):
            author = etree.Element("author")
            author.text = a
            authors.append(author)
        meta.append(authors)

        lib_info_fname = self.find_in_path("{0}_meta.xml".format(self._name))
        if lib_info_fname:
            xi = self.xi_include(os.path.basename(lib_info_fname))
            meta.append(xi)

        self._lib.append(meta)

    def __str__(self):
        xml = '<?xml version="1.0" encoding="UTF-8"?>\n'
        return xml + etree.tostring(self._lib, pretty_print=True)

    def add_to_others(self, fname, **kwargs):
        self.add_to_cat('others', fname, **kwargs)

    def add_to_cat(self, cat_name, fname, **kwargs):
        # first time add new category
        if cat_name not in self._cats:
            c = etree.Element('category', name=cat_name)
            cat_info_path = self.find_in_path("{0}_category_{1}.xml".format(self._name, cat_name))
            if cat_info_path:
                logging.info("adding: {0:s}".format(cat_info_path))
                cat_element = self.xi_include(os.path.basename(cat_info_path))
                c.append(cat_element)

            self._cats[cat_name] = c
            self._lib.append(c)
            self._cat_entries[cat_name] = {}

        # check for already added object
        if kwargs['name'] in self._cat_entries[cat_name]:
            logging.warning('object %s already in library', kwargs['name'])
            return

        self._cat_entries[cat_name][kwargs['name']] = True

        entry = etree.Element('entry', **kwargs)
        entry.append(self.xi_include(os.path.basename(fname)))
        self._cats[cat_name].append(entry)

    def xi_include(self, fname):
        return etree.Element('{http://www.w3.org/2001/XInclude}include', href=fname, parse="xml")

    def sort_cat(self, cat):
        entries = cat.findall('entry')
        sorted_entries = sorted(entries, key=lambda x: x.get('name'))

        for child in entries:
            cat.remove(child)

        for child in reversed(sorted_entries):
            cat.insert(0, child)

    def sort(self):
        cats = self._lib.findall('category')
        sorted_cats = sorted(cats, key=lambda x: x.get('name'))

        for child in cats:
            self._lib.remove(child)

        for child in reversed(sorted_cats):
            self.sort_cat(child)
            self._lib.insert(0, child)



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2016 by Serge Poltavski                                   #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from docobjectvisitor import DocObjectVisitor
from pd.coregui import Color
from pd.comment import Comment
from pd.factory import make_by_name
from pddoc.pdpage import PdPage
from pd.message import Message
from pd.obj import PdObject
from pd.parser import Parser
from pddoc.docobject import DocPar
import os


def add_text_dot(string):
    string = string.strip()
    if len(string) == 0:
        return string

    if string[-1] == '.':
        return string
    else:
        return string + '.'


def remove_text_dot(string):
    string = string.strip()
    if len(string) == 0:
        return string

    if string[-1] == '.':
        return string[0:-1]
    else:
        return string


class PdDocVisitor(DocObjectVisitor):
    PD_WINDOW_WIDTH = 715
    PD_WINDOW_HEIGHT = 555
    PD_HEADER_HEIGHT = 40
    PD_HEADER_FONT_SIZE = 20
    PD_HEADER_COLOR = Color(0, 255, 255)
    PD_HEADER_BG_COLOR = Color(100, 100, 100)
    PD_EXAMPLE_YOFFSET = 30
    PD_FOOTER_HEIGHT = 48
    PD_FOOTER_COLOR = Color(180, 180, 180)
    PD_INFO_WINDOW_WIDTH = 400
    PD_INFO_WINDOW_HEIGHT = 290
    PD_XLET_INDX_XPOS = 120
    PD_XLET_TYPE_XPOS = 150
    PD_XLET_INFO_XPOS = 240
    PD_ARG_NAME_COLOR = Color(240, 250, 250)

    def __init__(self):
        DocObjectVisitor.__init__(self)
        self.current_yoff = 0
        self._pp = PdPage("obj", self.PD_WINDOW_WIDTH, self.PD_WINDOW_HEIGHT)

    def meta_end(self, meta):
        self.add_header()
        self.current_yoff += self.PD_HEADER_HEIGHT
        self.current_yoff += 30

    def description_begin(self, d):
        super(self.__class__, self).description_begin(d)
        self._pp.add_description(self._description, self.PD_HEADER_HEIGHT + 10)

    def pdascii_begin(self, t):
        cnv = super(self.__class__, self).pdascii_begin(t)
        self.copy_canvas_objects(cnv)
        self.copy_canvas_connections(cnv)

        self.current_yoff += cnv.height
        self.current_yoff += self.PD_EXAMPLE_YOFFSET

    def copy_canvas_connections(self, cnv):
        for conn in cnv.connections.values():
            self._pp.canvas.add_connection(conn[0].id, conn[1], conn[2].id, conn[3])

    def copy_canvas_objects(self, cnv):
        for obj in cnv.objects:
            obj.y += self.current_yoff
            obj.x += self.PD_EXAMPLE_YOFFSET
            self._pp.append_object(obj)

    def pdinclude_begin(self, t):
        db_path = '{0}.db'.format(self._library)
        PdObject.xlet_calculator.add_db(db_path)

        pd_parser = Parser()
        pd_parser.parse(t.file())
        cnv = pd_parser.canvas
        self.copy_canvas_objects(cnv)
        self.copy_canvas_connections(cnv)
        self.current_yoff += cnv.height

    def methods_begin(self, m):
        self.add_section("methods:")

    def methods_end(self, m):
        self.current_yoff += 10

    def method_begin(self, m):
        msg_atoms = [m.name()]
        # for i in m.items():
            # msg_atoms.append(i.param_name())

        msg = Message(self.PD_XLET_INDX_XPOS, self.current_yoff, msg_atoms)
        msg.calc_brect()
        self._pp.append_object(msg)

        info_text = m.text().strip()
        info_text = add_text_dot(info_text)

        if len(m.items()) > 0:
            info_text += " Arguments are: "

        info = []
        info.append(self._pp.add_txt(info_text, self.PD_XLET_INFO_XPOS, self.current_yoff))

        # add method arguments
        for i in m.items():
            param_name = i.param_name()
            arg_descr = param_name

            if i.text() is not None:
                arg_descr += ": " + i.text().strip()

            if i.type():
                arg_descr = "{0} Type: {1}. ".format(add_text_dot(arg_descr), i.type())


            value_range = self.format_range(i)
            if len(value_range) > 0:
                # add dot after description
                arg_descr = add_text_dot(arg_descr)

                arg_descr += " " + value_range

            if len(i.enum()) > 0:
                arg_descr = "{0} Allowed values: {1}.".format(add_text_dot(arg_descr), ', '.join(i.enum()))

            hl_text = self._pp.make_txt(param_name, 0, 0)
            hl_text.calc_brect()
            bg = self._pp.make_background(0, 0, hl_text.width + 8, hl_text.height + 8, color=self.PD_ARG_NAME_COLOR)
            self._pp.append_object(bg)

            txt_obj = self._pp.add_txt(arg_descr, self.PD_XLET_INFO_XPOS + 10, self.current_yoff)
            setattr(txt_obj, 'background_obj', bg)

            info.append(txt_obj)

        self._pp.place_in_col(info, self.current_yoff, 15)

        for obj in info:
            if hasattr(obj, 'background_obj'):
                bg = getattr(obj, 'background_obj')
                bg.x = obj.x
                bg.y = obj.y

        info.append(msg)
        br = self._pp.group_brect(info)
        self.current_yoff += br[3] + 15

    def inlets_begin(self, inlets):
        super(self.__class__, self).inlets_begin(inlets)
        self.add_section("inlets:")
        inlets.enumerate()

    def inlets_end(self, inlets):
        self.current_yoff += 10

        if inlets.is_empty():
            self.current_yoff += 10

    def inlet_begin(self, inlet):
        self._pp.add_txt("{0}.".format(inlet.number()), self.PD_XLET_INDX_XPOS, self.current_yoff)

    def xinfo_begin(self, xinfo):
        tlist = []
        if xinfo.on():
            t1 = self._pp.add_txt("*{0}*".format(xinfo.on()), self.PD_XLET_TYPE_XPOS, self.current_yoff)
            tlist.append(t1)

        rng = self.format_range(xinfo)
        t2 = self._pp.add_txt("{0}. {1}".format(xinfo.text(), rng), self.PD_XLET_INFO_XPOS, self.current_yoff)
        tlist.append(t2)

        _, _, _, h = self._pp.group_brect(tlist)
        self.current_yoff += h + 5

    def outlets_begin(self, outlets):
        super(self.__class__, self).outlets_begin(outlets)
        self.add_section("outlets:")
        outlets.enumerate()

    def outlets_end(self, outlets):
        self.current_yoff += 10

        if outlets.is_empty():
            self.current_yoff += 10

    def outlet_begin(self, outlet):
        y = self.current_yoff
        t1 = self._pp.add_txt("{0}.".format(outlet.number()), self.PD_XLET_INDX_XPOS, y)
        t2 = self._pp.add_txt(outlet.text(), self.PD_XLET_INFO_XPOS, y)

        _, _, _, h = self._pp.group_brect([t1, t2])
        self.current_yoff += h + 5

    def arguments_begin(self, args):
        super(self.__class__, self).arguments_begin(args)
        self.add_section("arguments:")
        args.enumerate()

    def arguments_end(self, outlets):
        self.current_yoff += 10

    def properties_begin(self, p):
        super(self.__class__, self).properties_begin(p)
        self.add_section("properties:")

    def property_begin(self, m):
        props = list()

        # add message with property name
        prop_get_name = "{0}".format(m.name())
        props.append(Message(self.PD_XLET_INDX_XPOS, self.current_yoff, [prop_get_name]))
        self._pp.append_list(props)

        # props description
        prop_descr = ""
        if not (m.is_alias() or m.is_flag()):
            if m.readonly():
                prop_descr += "(readonly) Get "
            else:
                prop_descr += "Get/Set "

        prop_descr += m.text()

        if m.type() and not (m.is_alias() or m.is_flag()):
            prop_descr = "{0} Type: {1}. ".format(add_text_dot(prop_descr), m.type())

        if m.units():
            prop_descr = "{0} Units: {1}. ".format(add_text_dot(prop_descr), m.units())

        if m.default():
            prop_descr = "{0} Default value: {1}. ".format(add_text_dot(prop_descr), m.default())

        if m.min() and m.max():
            prop_descr = "{0} Value range: {1}-{2}. ".format(add_text_dot(prop_descr), m.min(), m.max())
        elif m.min():
            prop_descr = "{0} Min value: {1}. ".format(add_text_dot(prop_descr), m.min())
        elif m.max():
            prop_descr = "{0} Max value: {1}. ".format(add_text_dot(prop_descr), m.max())

        if len(m.enum()) > 0:
            prop_descr = "{0} Allowed values: {1}.".format(add_text_dot(prop_descr), ', '.join(m.enum()))

        info = list()
        info.append(self._pp.add_txt(remove_text_dot(prop_descr), self.PD_XLET_INFO_XPOS, self.current_yoff))

        self._pp.place_in_col(info, self.current_yoff, 15)
        br = self._pp.group_brect(info + props)
        self.current_yoff += br[3] + 10

    def properties_end(self, p):
        self.current_yoff += 10

    def info_begin(self, info):
        lst = []
        for p in info.items():
            if isinstance(p, DocPar):
                t = self._pp.make_txt(p.text(), self.PD_XLET_INFO_XPOS - 30, 0)
                lst.append(t)

        self._pp.place_in_col(lst, self.PD_HEADER_HEIGHT + 40, 10)
        brect = self._pp.group_brect(lst)
        bg = self._pp.make_background(brect[0] - 5, brect[1],
                                      self._pp.width - self.PD_XLET_INFO_XPOS + 15,
                                      brect[3] + 20, Color(250, 250, 250))
        self._pp.append_object(bg)
        self._pp.append_list(lst)
        self.current_yoff = brect[1] + brect[3]

    def argument_begin(self, arg):
        y = self.current_yoff
        super(self.__class__, self).argument_begin(arg)

        t1 = self._pp.add_txt("{0}.".format(arg.number()), self.PD_XLET_INDX_XPOS, y)
        t2 = self._pp.add_txt(arg.type(), self.PD_XLET_TYPE_XPOS, y)
        self.add_background_for_txt(arg.main_info_prefix(),
                                    self.PD_XLET_INFO_XPOS,
                                    y,
                                    self.PD_ARG_NAME_COLOR)

        rng = self.format_range(arg)
        t3 = self._pp.add_txt("{0}. {1}".format(arg.main_info(), rng), self.PD_XLET_INFO_XPOS, y)
        _, _, _, h = self._pp.group_brect([t1, t2, t3])
        self.current_yoff += h + 5

    def object_end(self, obj):
        LNK_Y = 45
        # index link
        idx_lnk = self._pp.make_link(0, LNK_Y, "../index-help.pd", "index")
        del1 = self._pp.make_txt("::", 0, LNK_Y)
        # library link
        lib_lnk = self._pp.make_link(0, LNK_Y,
                                     "{0}-help.pd".format(self._library),
                                     "{0}".format(self._library))
        del2 = self._pp.make_txt("::", 0, LNK_Y)
        # category link
        cat_lnk = self._pp.make_link(0, LNK_Y,
                                     "{0}.{1}-help.pd".format(self._library, self._category),
                                     "{0}".format(self._category))
        menu = [idx_lnk, del1, lib_lnk, del2, cat_lnk]
        self._pp.place_in_row(menu, 10, 7)
        self._pp.append_list(menu)
        self.add_footer()

    def render(self):
        return self._pp.to_string()

    @classmethod
    def format_range(cls, obj):
        r = obj.range()
        if len(r) == 2:
            if not r[0]:
                return "Max value: {0}".format(r[1])
            if not r[1]:
                return "Min value: {0}".format(r[0])

            return "Range: {0}...{1}".format(r[0], r[1])
        else:
            return ""

    def add_section(self, txt):
        hr = self._pp.make_styled_hrule(self.current_yoff)
        self._pp.append_object(hr)
        lbl = self._pp.make_section_label(self.current_yoff + 5, txt, font_size=14)
        self.current_yoff += 10
        self._pp.append_object(lbl)

    def add_header(self):
        lbl = self.add_header_label()
        if not self._is_gui:
            self.add_header_example_object(lbl, self._title)

    def add_header_example_object(self, lbl, title):
        seq = []
        for a in filter(lambda n: n != title, map(lambda i: i["name"], self._aliases)) + [title]:
            seq.append(make_by_name(a))

        self._pp.place_in_row(seq, 0, 20)
        _, _, w, h = self._pp.group_brect(seq)
        y = (lbl.height - h) / 2
        x = (lbl.width - w) - 20
        self._pp.move_to_x(seq, x)
        self._pp.move_to_y(seq, y)
        self._pp.append_list(seq)

    def add_header_label(self):
        return self._pp.add_header("{0}".format(self._title))

    def add_footer(self):
        ft = self._pp.make_footer(self.current_yoff + 20, height=self.PD_FOOTER_HEIGHT)
        y = ft.y
        self._pp.append_object(ft)
        self.add_footer_library(y)
        self.add_also(y + 15)
        self.add_footer_more_info(y)

    def add_footer_more_info(self, y):
        # more info
        pd = self._pp.make_subpatch('info', 10, y + 22, self.PD_INFO_WINDOW_WIDTH, self.PD_INFO_WINDOW_HEIGHT)

        def add_subpatch_text(x, y, txt):
            self._pp.add_subpatch_txt('info', txt, x, y)

        bg = self._pp.make_background(1, 1, 107, self.PD_INFO_WINDOW_HEIGHT - 3, self.PD_FOOTER_COLOR)
        self._pp.add_subpatch_obj('info', bg)
        xc1 = 10
        xc2 = 120
        yrows = range(10, 300, 22)
        row = 0
        add_subpatch_text(xc1, yrows[row], "library:")
        add_subpatch_text(xc2, yrows[row], self._library)
        row += 1
        add_subpatch_text(xc1, yrows[row], "version:")
        add_subpatch_text(xc2, yrows[row], self._version)
        row += 1
        add_subpatch_text(xc1, yrows[row], "object:")
        add_subpatch_text(xc2, yrows[row], self._title)
        row += 1
        add_subpatch_text(xc1, yrows[row], "category:")
        add_subpatch_text(xc2, yrows[row], self._category)
        row += 1

        if self._since:
            add_subpatch_text(xc1, yrows[row], "since:")
            add_subpatch_text(xc2, yrows[row], self._since)
            row += 1

        add_subpatch_text(xc1, yrows[row], "authors:")
        add_subpatch_text(xc2, yrows[row], ", ".join(self._authors))
        row += 1
        add_subpatch_text(xc1, yrows[row], "license:")
        if not self._license.get('url', ' '):
            add_subpatch_text(xc2, yrows[row], self._license['name'])
        else:
            lnk = self._pp.make_link(xc2, yrows[row], self._license['url'], self._license['name'])
            pd.append_object(lnk)
        row += 1
        if self._keywords:
            add_subpatch_text(xc1, yrows[row], "keywords:")
            add_subpatch_text(xc2, yrows[row], ", ".join(self._keywords))
            row += 1
        if self._website:
            add_subpatch_text(xc1, yrows[row], "website:")
            lnk = self._pp.make_link(xc2, yrows[row], self._website, self._website)
            pd.append_object(lnk)
            row += 1
        if self._contacts:
            add_subpatch_text(xc1, yrows[row], "contacts:")
            add_subpatch_text(xc2, yrows[row], self._contacts)
            row += 1

        pd.append_object(PdObject("declare", x=xc2, y=yrows[row], args=["-lib", "ceammc", "-lib", "creammc"]))

        ypos = self.PD_INFO_WINDOW_HEIGHT - 22
        delim = self._pp.make_hrule(xc2, ypos, width=270)
        pd.append_object(delim)
        add_subpatch_text(xc2, ypos, "generated by pddoc")
        self._pp.canvas.append_subpatch(pd)

    def add_footer_library(self, y):
        # library:
        self._pp.add_txt("library: {0} v{1}".format(self._library, self._version), 10, y + 3)

    def add_background_for_txt(self, txt, x, y, color, **kwargs):
        if len(txt) < 1:
            return

        padx = kwargs.get("padx", 5)
        pady = kwargs.get("pady", 5)

        c = Comment(x, y, txt.split(' '))
        c.calc_brect()
        bg = self._pp.make_background(x + 1, y + 1, c.width + padx, c.height + pady, color)
        self._pp.append_object(bg)

    def add_also(self, y):
        if len(self._see_also) < 1:
            return

        # see also:
        label = self._pp.make_txt("see also:", 0, 0)
        also_objects = [label]
        for see in self._see_also:
            if 'is_link' in see and see['is_link'] == True:
                lnk = self._pp.make_link(0, 0,
                                         "{0}-help.pd".format(see['name']),
                                         "[{0}]".format(see['name']))
                lnk.set_bg_color(Color(200, 200, 200))
                also_objects.append(lnk)
            else:
                also_objects.append(make_by_name(see['name']))

        self._pp.place_in_row(also_objects, 0, 10)
        _, _, w, h = self._pp.group_brect(also_objects)
        self._pp.move_to_y(also_objects, y)
        self._pp.move_to_x(also_objects, (self._pp.width - w) - 20)
        self._pp.append_list(also_objects)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

import imp
import logging

from xletdatabase import XletDatabase


class XletCalcDatabase(XletDatabase):
    _counter = 0

    def __init__(self, path, extname):
        super(XletCalcDatabase, self).__init__(extname)
        self._extname = extname
        XletCalcDatabase._counter += 1
        self._module = None

        try:
            self._module = imp.load_source("plugin{0:d}".format(XletCalcDatabase._counter), path)
        except IOError as e:
            logging.error("Plugin not found: {0:s}".format(path))
            raise e

    def outlets(self, name, args=None):
        if args is None:
            args = []
        return self._module.outlets(name, args)

    def inlets(self, name, args=None):
        if args is None:
            args = []
        return self._module.inlets(name, args)

    def has_object(self, name):
        return self._module.has_object(name)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2016 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from lxml import etree
import os
import logging


def get_parser():
    schema_root = etree.parse(os.path.join(os.path.dirname(__file__), 'share', 'pddoc.xsd')).getroot()
    schema = etree.XMLSchema(schema_root)
    return etree.XMLParser(schema=schema)


def parse_xml(path):
    if not os.path.exists(path):
        logging.error("File not exists: \"%s\"", path)
        return None

    try:
        xml = etree.parse(path, get_parser())
        xml.xinclude()
    except etree.XMLSyntaxError as e:
        logging.error("XML syntax error:\n \"%s\"\n\twhile parsing file: \"%s\"", e, path)
        return None

    pddoc = xml.getroot()
    for obj in pddoc.findall('object'):
        fix_section_order(obj)

    return xml


def fix_section_order(obj_tag):
    sorted_children = sorted(obj_tag, key=lambda n: ('title',
                                                     'meta', 'info',
                                                     'example', 'arguments', 'properties', 'methods',
                                                     'inlets', 'outlets').index(n.tag))
    for child in obj_tag:
        obj_tag.remove(child)

    for child in reversed(sorted_children):
        obj_tag.insert(0, child)




#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2014 by Serge Poltavski                                   #
# serge.poltavski@gmail.com                                               #
# #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

from __future__ import print_function

__author__ = 'Serge Poltavski'

import cairo
import textwrap
from math import pi
import logging

from .pdpainter import PdPainter
from .pd import XLET_GUI, XLET_SOUND, XLET_MESSAGE, Canvas


class CairoPainter(PdPainter):
    def __init__(self, width, height, output, fmt="png", **kwargs):
        PdPainter.__init__(self)

        from pd.drawstyle import DrawStyle
        self.style = DrawStyle()

        self.st_font_slant = cairo.FONT_SLANT_NORMAL
        self.st_font_weight = cairo.FONT_WEIGHT_NORMAL
        self.st_line_join = cairo.LINE_JOIN_ROUND

        assert width > 0
        assert height > 0

        self.width = width
        self.height = height

        assert fmt
        fmt = fmt.lower()
        self.format = fmt
        self.output = output

        if fmt == 'png':
            self.ims = cairo.ImageSurface(cairo.FORMAT_ARGB32, self.width, self.height)
        elif fmt == 'pdf':
            assert self.output
            self.ims = cairo.PDFSurface(self.output, self.width, self.height)
        elif fmt == 'svg':
            assert self.output
            self.ims = cairo.SVGSurface(self.output, self.width, self.height)
        else:
            msg = "unsupported format: \"{0:s}\"".format(fmt)
            logging.error(msg)
            raise RuntimeError(msg)

        self.cr = cairo.Context(self.ims)

        self.set_src_color(self.style.fill_color)
        self.cr.rectangle(0, 0, self.width, self.height)
        self.cr.fill()

        self.cr.select_font_face(self.style.font_family, self.st_font_slant, self.st_font_weight)
        self.cr.set_font_size(self.style.font_size)
        self.cr.set_line_join(self.st_line_join)

        matrix = cairo.Matrix()
        if 'xoffset' in kwargs:
            matrix.translate(kwargs['xoffset'], 0)
        if 'yoffset' in kwargs:
            matrix.translate(0, kwargs['yoffset'])

        self.cr.set_matrix(matrix)

    def __del__(self):
        if self.format == 'png' and self.output:
            self.ims.write_to_png(self.output)

    def draw_rect(self, x, y, w, h, **kwargs):
        self.cr.save()

        if 'width' in kwargs:
            self.cr.set_line_width(kwargs['width'])

        if self.cr.get_line_width() % 2 == 1:
            x = round(x) + 0.5
            y = round(y) + 0.5

        self.cr.rectangle(x, y, w, h)

        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])

        self.cr.stroke()
        self.cr.restore()

    def set_src_color(self, rgb):
        if isinstance(rgb, float):
            self.cr.set_source_rgb(rgb, rgb, rgb)
        elif len(rgb) == 3:
            self.cr.set_source_rgb(rgb[0], rgb[1], rgb[2])

    def draw_box(self, x, y, w, h):
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.set_src_color(self.style.obj_border_color)
        self.cr.rectangle(x + 0.5, y + 0.5, w, h)
        self.cr.stroke_preserve()
        self.set_src_color(self.style.obj_fill_color)
        self.cr.fill()
        self.cr.restore()

    def draw_txt(self, x, y, txt):
        ascent, descent, height, ax, ay = self.cr.font_extents()
        self.cr.save()
        self.set_src_color(self.style.obj_text_color)
        self.cr.move_to(x, y + ascent)
        self.cr.show_text(txt)
        self.cr.restore()

    def draw_xlets(self, xlets, x, y, obj_width):
        if not xlets:
            return

        inlet_space = 0
        if len(xlets) > 1:
            inlet_space = (obj_width - len(xlets) * self.style.xlet_width) / (len(xlets) - 1)

        for num in range(0, len(xlets)):
            inx = x + num * inlet_space
            if num != 0:
                inx += num * self.style.xlet_width

            xlet = xlets[num]

            if xlet in (XLET_MESSAGE, XLET_GUI):
                self.set_src_color(self.style.xlet_msg_color)
            else:
                self.set_src_color(self.style.xlet_snd_color)

            iny = y
            self.cr.set_line_width(self.style.obj_line_width)
            if self.style.obj_line_width % 2 == 1:
                inx = round(inx) + 0.5
                iny = round(iny) + 0.5

            self.cr.rectangle(inx, iny, self.style.xlet_width, self.xlet_height(xlet))
            self.cr.stroke_preserve()

            if xlet == XLET_SOUND:
                self.cr.fill()
                self.cr.stroke()
            else:
                self.cr.stroke()

            num += 1

    def draw_subpatch(self, subpatch):
        if subpatch.is_graph_on_parent():
            x = subpatch.x
            y = subpatch.y
            w = subpatch._gop['width']
            h = subpatch._gop['height']
            self.draw_rect(x, y, w, h, color=self.style.obj_border_color, width=1)
            m = self.cr.get_matrix()
            m.translate(x - subpatch._gop['xoff'], y - subpatch._gop['yoff'])
            self.cr.save()
            self.cr.set_matrix(m)
            for obj in subpatch.objects:
                if obj.draw_on_parent():
                    obj.draw(self)

            self.cr.restore()

            if not subpatch._gop['hide_args']:
                txt = "pd " + subpatch.args_to_string()
                self.draw_text(x + self.style.obj_pad_x,
                               y + self.style.font_size,
                               txt,
                               font_size=self.style.font_size)

            self.draw_xlets(subpatch.inlets(), x, y, w)
            self.draw_xlets(subpatch.outlets(), x, y + h - self.style.xlet_msg_height, w)
            return
        else:
            self.draw_object_txt(subpatch, "pd " + subpatch.args_to_string())

    def draw_gop(self, obj, gop_canvas):
        x, y, w, h = gop_canvas.gop_rect()
        self.cr.save()
        self.draw_box(obj.x, obj.y, w, h)
        self.cr.translate(obj.x - x, obj.y - y)
        gop_canvas.draw_gop(self)
        self.cr.restore()
        pass

    def text_size(self, txt, font_size=None):
        self.cr.save()
        self.cr.select_font_face(self.style.font_family)

        if font_size is not None:
            self.cr.set_font_size(int(font_size))
        else:
            self.cr.set_font_size(self.style.font_size)

        w = self.cr.text_extents(txt)[2]
        h = self.cr.font_extents()[2]
        self.cr.restore()
        return w, h

    def box_size(self, txt):
        w, h = self.text_size(txt)
        # box padding
        w += self.style.obj_pad_x * 3  # set by empirical way
        h += self.style.obj_pad_y * 2

        # objects have minimal width. otherwise such objects as [*] could be too narrow
        w = max(w, self.style.obj_min_width)
        return w, h

    def draw_hightlight(self, x, y, w, h):
        self.cr.save()
        pad = self.style.highlight_padding
        self.cr.rectangle(x + 0.5 - pad, y + 0.5 - pad, w + pad * 2, h + pad * 2)
        self.set_src_color(self.style.highlight_color)
        self.cr.fill()
        self.cr.restore()

    def draw_object(self, obj):
        self.draw_object_txt(obj, obj.to_string())

    def draw_object_txt(self, obj, txt):
        self.cr.save()
        self.cr.select_font_face(self.style.font_family)

        x = obj.x
        y = obj.y
        w, h = self.box_size(txt)

        obj.set_width(w)
        obj.set_height(h)

        if hasattr(obj, "highlight"):
            self.draw_hightlight(x, y, w, h)

        self.draw_box(x, y, w, h)
        self.draw_txt(x + self.style.obj_pad_x, y + self.style.obj_pad_y + 1, txt)
        self.draw_xlets(obj.inlets(), x, y, w)
        self.draw_xlets(obj.outlets(), x, y + h - self.style.xlet_msg_height, w)
        self.cr.restore()

    def draw_message_box(self, x, y, w, h):
        edge_w = 8
        edge_h = 10

        self.cr.save()
        self.set_src_color(self.style.obj_border_color)
        self.cr.set_line_width(self.style.obj_line_width)
        tx = x + 0.5
        ty = y + 0.5
        self.cr.move_to(tx, ty)
        self.cr.line_to(tx, ty + h)
        self.cr.line_to(tx + w, ty + h)
        self.cr.line_to(tx + w - edge_w / 2.0, ty + (h + edge_h) / 2.0)
        self.cr.line_to(tx + w - edge_w / 2.0, ty + (h - edge_h) / 2.0)
        self.cr.line_to(tx + w, ty)
        self.cr.line_to(tx, ty)
        self.cr.stroke_preserve()
        self.set_src_color(self.style.msg_fill_color)
        self.cr.fill()
        self.cr.stroke()
        self.cr.restore()

    def message_size(self, txt):
        w, h = self.box_size(txt)
        # increase width for message rightmost curved tail: [msg(
        w += self.style.obj_pad_y * 2
        return w, h

    def draw_message(self, message):
        txt = message.to_string()
        x = message.x
        y = message.y
        w, h = self.message_size(txt)

        message.set_height(h)
        message.set_width(w)

        self.draw_message_box(x, y, w, h)
        self.draw_txt(x + self.style.obj_pad_x, y + self.style.obj_pad_y + 1, txt)
        self.draw_xlets(message.inlets(), x, y, w)
        self.draw_xlets(message.outlets(), x, y + h - self.style.xlet_msg_height, w)

    def xlet_height(self, t):
        if t == XLET_GUI:
            return self.style.xlet_gui_height
        else:
            return self.style.xlet_msg_height

    def inlet_connection_coords(self, obj, inlet_no):
        inlets = obj.inlets()

        xlet_space = 0
        if len(inlets) > 1:
            xlet_space = (obj.width - len(inlets) * self.style.xlet_width) / float(len(inlets) - 1)

        x = obj.x + (xlet_space + self.style.xlet_width) * inlet_no + self.style.xlet_width / 2.0
        y = obj.y

        return x, y

    def outlet_connection_coords(self, obj, outlet_no):
        outlets = obj.outlets()

        xlet_space = 0
        if len(outlets) > 1:
            xlet_space = (obj.width - len(outlets) * self.style.xlet_width) / float(len(outlets) - 1)

        x = obj.x + (xlet_space + self.style.xlet_width) * outlet_no + self.style.xlet_width / 2.0
        if isinstance(obj, Canvas):
            if obj.is_graph_on_parent():
                y = obj._gop['height'] + obj.y
            else:
                y = obj.top + self.style.obj_height
        else:
            y = obj.bottom
        return x, y

    def draw_connections(self, canvas):
        for key, c in canvas.connections.items():
            src_obj = c[0]
            src_outl = c[1]
            if not src_obj.outlets():
                continue

            src_outl_type = src_obj.outlets()[src_outl]
            dest_obj = c[2]
            dest_inl = c[3]
            if not dest_obj.inlets():
                continue

            dest_inl_type = dest_obj.inlets()[dest_inl]

            sx, sy = self.outlet_connection_coords(src_obj, src_outl)
            dx, dy = self.inlet_connection_coords(dest_obj, dest_inl)

            # print sx, sy, "->", dx, dy

            self.cr.save()
            if dest_inl_type == XLET_SOUND and src_outl_type == XLET_SOUND:
                self.cr.set_line_width(self.style.conn_snd_width)
                self.set_src_color(self.style.conn_snd_color)
                # pixel correction
                if self.style.conn_snd_width % 2 == 0:
                    sx += 0.5
                    dx += 0.5

                self.cr.move_to(sx, sy)
                self.cr.line_to(dx, dy)
                self.cr.stroke()
                self.cr.set_dash(self.style.conn_snd_dash)
                self.set_src_color(self.style.conn_snd_color2)
            else:
                self.cr.set_line_width(self.style.conn_msg_width)
                self.set_src_color(self.style.conn_msg_color)
                # pixel correction
                if self.style.conn_msg_width % 2 == 1 and (sx - int(sx) == 0):
                    sx += 0.5
                    dx += 0.5

                sy += self.style.conn_msg_width

            self.cr.move_to(sx, sy)
            self.cr.line_to(dx, dy)
            self.cr.stroke()
            self.cr.restore()

        pass

    def draw_comment(self, comment):
        txt = comment.text()
        lines = textwrap.wrap(txt, 59)

        self.set_src_color(self.style.comment_color)
        line_height = self.cr.font_extents()[2]
        lnum = 0
        for line in lines:
            line = ";\n".join(line.split(";")).split("\n")
            for subl in line:
                subl = subl.strip()
                self.draw_txt(comment.x, comment.y + lnum * line_height, subl)
                lnum += 1

    def draw_poly(self, vertexes, **kwargs):
        assert len(vertexes) > 1
        assert len(vertexes[0]) == 2
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.cr.move_to(vertexes[0][0] + 0.5, vertexes[0][1] + 0.5)
        for v in vertexes[1:]:
            self.cr.rel_line_to(v[0], v[1])

        self.cr.line_to(vertexes[0][0] + 0.5, vertexes[0][1] + 0.5)

        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        if 'outline' in kwargs:
            self.set_src_color(kwargs['outline'])
            self.cr.stroke()

        self.cr.restore()

    def draw_text(self, x, y, text, **kwargs):
        self.cr.save()

        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])
        else:
            self.set_src_color((0, 0, 0))

        if 'font_size' in kwargs:
            self.cr.set_font_size(kwargs['font_size'])

        if 'font' in kwargs:
            self.cr.select_font_face(kwargs['font'])

        self.cr.move_to(x, y)
        self.cr.show_text(text)
        self.cr.restore()

    def draw_inlets(self, inlets, x, y, width):
        self.draw_xlets(inlets, x, y, width)

    def draw_outlets(self, outlets, x, y, width):
        if not outlets:
            return

        yoffset = self.style.xlet_height(outlets[0])
        self.draw_xlets(outlets, x, y - yoffset, width)

    def draw_arc(self, x, y, radius, start_angle, end_angle, **kwargs):
        self.cr.save()
        if 'width' in kwargs:
            self.cr.set_line_width(float(kwargs['width']))

        self.cr.move_to(x, y)
        self.cr.arc(x, y, radius, start_angle, end_angle)

        if 'outline' in kwargs:
            self.set_src_color(kwargs['outline'])

        self.cr.stroke()
        self.cr.restore()

    def draw_circle(self, x, y, radius, **kwargs):
        self.cr.save()
        self.cr.set_line_width(self.style.obj_line_width)
        self.cr.arc(x, y, radius, 0, 2 * pi)
        if 'fill' in kwargs:
            self.set_src_color(kwargs['fill'])
            self.cr.fill_preserve()

        if 'outline' in kwargs:
            self.set_src_color(kwargs['outline'])
            self.cr.stroke()

        self.cr.restore()

    def draw_line(self, x0, y0, x1, y1, **kwargs):
        self.cr.save()
        self.cr.set_line_width(float(kwargs.get('width', 1)))
        if 'color' in kwargs:
            self.set_src_color(kwargs['color'])

        self.cr.set_line_cap(cairo.LINE_CAP_BUTT)
        if int(self.cr.get_line_width()) % 2 == 1:
            y0 = round(y0) + 0.5
            y1 = round(y1) + 0.5

        if self.cr.get_line_width() == 1:
            x0 = round(x0) + 0.5
            x1 = round(x1) + 0.5

        self.cr.move_to(x0, y0)
        self.cr.line_to(x1, y1)
        self.cr.stroke()
        self.cr.restore()

    def draw_graph(self, graph):
        print(graph)



#!/usr/bin/env python
# coding=utf-8

#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

 
__author__ = 'Serge Poltavski'

from xletdatabase import XletDatabase
import os
import re
import logging

from . import XLET_MESSAGE, XLET_SOUND


class ObjectRecord(object):
    def __init__(self, path, inlets=None, outlets=None):
        self._path = path
        self._inlets = inlets
        self._outlets = outlets

    @property
    def inlets(self):
        return self._inlets

    @inlets.setter
    def inlets(self, inlets):
        self._inlets = inlets

    @property
    def outlets(self):
        return self._outlets

    @outlets.setter
    def outlets(self, outlets):
        self._outlets = outlets

    @property
    def path(self):
        return self._path

    def no_info(self):
        return self._inlets is None or self._outlets is None

    def __str__(self):
        return "Object: {0:s}\n\tinlets: {1:s}\n\toutlets: {2:s}".format(self._path, self._inlets, self._outlets)


class XletPatchLookup(XletDatabase):
    lines_re = re.compile("(#(.*?)[^\\\])\r?\n?;\r?\n", re.MULTILINE | re.DOTALL)
    split_re = re.compile(" |\r\n?|\n", re.MULTILINE)

    def __init__(self, dirs=None):
        XletDatabase.__init__(self, "pdpatch")
        if dirs is None:
            dirs = []
        self._dirs = dirs
        self._dirs.append(os.getcwd())
        self._cache = {}

    def clean_name(self, name):
        return re.sub("/[\W]+/", '', name) + ".pd"

    def get_object(self, name):
        cleanName = self.clean_name(name)
        return self._cache.get(cleanName)

    def has_object(self, name):
        cleanName = self.clean_name(name)

        if cleanName in self._cache:
            return True

        for directory in self._dirs:
            path = os.path.join(directory, cleanName)
            if os.path.exists(path):
                self._cache[cleanName] = ObjectRecord(path)
                return True

        return False

    def inlets(self, name, args=None):
        if args is None:
            args = []
        cname = self.clean_name(name)
        if cname not in self._cache:
            return []

        if self._cache[cname].no_info():
            self.calc_xlets(cname)

        return self._cache[cname].inlets

    def outlets(self, name, args=None):
        if args is None:
            args = []
        cname = self.clean_name(name)
        if cname not in self._cache:
            return []

        if self._cache[cname].no_info():
            self.calc_xlets(cname)

        return self._cache[cname].outlets

    def calc_xlets(self, cname):
        assert cname in self._cache
        object_info = self._cache[cname]
        # assert isinstance(object_info, ObjectRecord)
        inlets_coords = []
        outlets_coords = []

        try:
            f = open(object_info.path, 'r')

            lines = f.read()
            subpatch_counter = -1
            for found in self.lines_re.finditer(lines):
                line = found.group(1)
                atoms = self.split_re.split(line)

                if atoms[0] == "#N" and atoms[1] == "canvas":
                    subpatch_counter += 1

                if subpatch_counter == 0:
                    if atoms[1] == "obj":
                        x = lambda: int(atoms[2])
                        if atoms[4] == "inlet":
                            inlets_coords.append((XLET_MESSAGE, x()))
                        if atoms[4] == "inlet~":
                            inlets_coords.append((XLET_SOUND, x()))
                        if atoms[4] == "outlet":
                            outlets_coords.append((XLET_MESSAGE, x()))
                        if atoms[4] == "outlet~":
                            outlets_coords.append((XLET_SOUND, x()))

                if atoms[0] == "#X" and atoms[1] == "restore":
                    subpatch_counter -= 1

        except IOError:
            logging.error("can't open file: {0:s}".format(object_info.path))

        inlets_coords.sort(key=lambda el: el[1])
        outlets_coords.sort(key=lambda el: el[1])
        object_info.inlets = [k for k, v in inlets_coords]
        object_info.outlets = [k for k, v in outlets_coords]



#!/usr/bin/env python
# coding=utf-8
import argparse
import shutil
import logging
import os
from pddoc.parser import parse_xml
from pddoc.pd import factory

from pddoc.htmldocvisitor import HtmlDocVisitor
from pddoc.docobject import DocObject
#   Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'


def main():
    arg_parser = argparse.ArgumentParser(description='PureData pddoc to html converter')
    arg_parser.add_argument('name', metavar='PDDOC', help="Documentation file in PDDOC format")
    arg_parser.add_argument('output', metavar='OUTNAME', nargs='?', default='',
                            help="HTML output file name")

    args = vars(arg_parser.parse_args())
    in_file = args['name']
    output = args['output']

    if not os.path.exists(in_file):
        logging.error("File not exists: \"%s\"", in_file)
        exit(1)

    if not output:
        output = os.path.splitext(os.path.basename(in_file))[0] + ".html"

    xml = parse_xml(in_file)
    if not xml:
        exit(1)

    css_file = "theme.css"
    factory.add_import('ceammc')

    pddoc = xml.getroot()
    for child_tag in pddoc:
        if child_tag.tag == "object":
            dobj = DocObject()
            dobj.from_xml(child_tag)

            v = HtmlDocVisitor()
            v.set_css_file(css_file)
            v.set_image_prefix(child_tag.attrib["name"])
            v.set_search_dir(os.path.dirname(in_file))

            # traverse doc
            dobj.traverse(v)

            # generate images
            v.generate_images()

            html_data = v.render()
            f = open(output, "w")
            f.write(html_data)
            f.close()

    # copy css theme file to current folder
    src_css = os.path.join(os.path.dirname(__file__), "../share", css_file)
    shutil.copyfile(src_css, css_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
#   serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

import ctypes
import cairo
import os
import cairo
import pango
import pangocairo

_initialized = False


def create_cairo_font_face_for_file(filename, faceindex=0, loadoptions=0):
    global _initialized
    global _freetype_so
    global _cairo_so
    global _ft_lib
    global _surface

    CAIRO_STATUS_SUCCESS = 0
    FT_Err_Ok = 0

    if not _initialized:
        # find shared objects
        _freetype_so = ctypes.CDLL("/opt/local/lib/libfreetype.dylib")
        _cairo_so = ctypes.CDLL("/opt/local/lib/libcairo.dylib")

        _cairo_so.cairo_ft_font_face_create_for_ft_face.restype = ctypes.c_void_p
        _cairo_so.cairo_ft_font_face_create_for_ft_face.argtypes = [ctypes.c_void_p, ctypes.c_int]
        _cairo_so.cairo_set_font_face.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
        _cairo_so.cairo_font_face_status.argtypes = [ctypes.c_void_p]
        _cairo_so.cairo_status.argtypes = [ctypes.c_void_p]

        # initialize freetype
        _ft_lib = ctypes.c_void_p()
        if FT_Err_Ok != _freetype_so.FT_Init_FreeType(ctypes.byref(_ft_lib)):
            raise "Error initialising FreeType library."

        class PycairoContext(ctypes.Structure):
            _fields_ = [("PyObject_HEAD", ctypes.c_byte * object.__basicsize__),
                        ("ctx", ctypes.c_void_p),
                        ("base", ctypes.c_void_p)]

        _surface = cairo.ImageSurface(cairo.FORMAT_A8, 0, 0)

        _initialized = True

    # create freetype face
    ft_face = ctypes.c_void_p()
    cairo_ctx = cairo.Context(_surface)
    cairo_t = PycairoContext.from_address(id(cairo_ctx)).ctx

    if FT_Err_Ok != _freetype_so.FT_New_Face(_ft_lib, filename, faceindex, ctypes.byref(ft_face)):
        raise Exception("Error creating FreeType font face for " + filename)

    # create cairo font face for freetype face
    cr_face = _cairo_so.cairo_ft_font_face_create_for_ft_face(ft_face, loadoptions)
    if CAIRO_STATUS_SUCCESS != _cairo_so.cairo_font_face_status(cr_face):
        raise Exception("Error creating cairo font face for " + filename)

    _cairo_so.cairo_set_font_face(cairo_t, cr_face)
    if CAIRO_STATUS_SUCCESS != _cairo_so.cairo_status(cairo_t):
        raise Exception("Error creating cairo font face for " + filename)

    face = cairo_ctx.get_font_face()

    return face


if __name__ == '__main__':
    share_path = os.path.join(os.path.dirname(__file__), "share", "terminus.ttf")
    face = create_cairo_font_face_for_file(share_path, 0)

    surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, 128, 128)
    ctx = cairo.Context(surface)
    ctx.set_font_face(face)
    ctx.set_font_size(14)
    ctx.move_to(0, 44)
    ctx.show_text("Hello,")
    ctx.move_to(30, 74)
    ctx.show_text("world!")

    del ctx

    surface.write_to_png("hello.png")



#!/usr/bin/env python
# coding=utf-8

# Copyright (C) 2015 by Serge Poltavski                                 #
# serge.poltavski@gmail.com                                             #
#                                                                         #
#   This program is free software; you can redistribute it and/or modify  #
#   it under the terms of the GNU General Public License as published by  #
#   the Free Software Foundation; either version 3 of the License, or     #
#   (at your option) any later version.                                   #
#                                                                         #
#   This program is distributed in the hope that it will be useful,       #
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        #
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         #
#   GNU General Public License for more details.                          #
#                                                                         #
#   You should have received a copy of the GNU General Public License     #
#   along with this program. If not, see <http://www.gnu.org/licenses/>   #

__author__ = 'Serge Poltavski'

from pddoc.pd import CoreGui
from pddoc.cairopainter import CairoPainter
from pddoc.pd import XLET_GUI
from math import pi, sin, cos


def create(atoms):
    assert len(atoms)
    return Knob.from_atoms(atoms)


# X obj 391 121
# name size  min max log init  send  receive  label lx ly  font font_size  bcolor   fcolor lcolor init_value steady
# knob 32 32 0   127 0   0     empty empty    empty 0  -8  0    8          -262144  -1     -1     0          1
class Knob(CoreGui):
    @staticmethod
    def from_atoms(atoms):
        return Knob(0, 0,
                    size=atoms[1],
                    min=atoms[3],
                    max=atoms[4],
                    log=atoms[5],
                    init=atoms[6],
                    send=atoms[7],
                    receive=atoms[8],
                    label=atoms[9],
                    label_xoff=atoms[10],
                    label_yoff=atoms[11],
                    font_type=atoms[12],
                    font_size=atoms[13],
                    bg_color=atoms[14],
                    fg_color=atoms[15],
                    label_color=atoms[16],
                    init_value=atoms[17],
                    steady=atoms[18])

    def __init__(self, x, y, **kwargs):
        CoreGui.__init__(self, "knob", x, y, [], **kwargs)
        self._width = float(kwargs.get("size", 32))
        self._height = self._width
        self._min = float(kwargs.get("min", 0))
        self._max = float(kwargs.get("max", 127))
        assert self._min <= self._max
        self._log = int(kwargs.get("log", 0))
        self._init = int(kwargs.get("init", 0))
        self._default_value = int(kwargs.get("default_value", 0))
        self._steady = int(kwargs.get("steady", 0))

    def inlets(self):
        return [XLET_GUI]

    def outlets(self):
        return [XLET_GUI]

    def draw(self, painter):
        assert isinstance(painter, CairoPainter)

        self.draw_label(painter)
        # self.draw_xlets(painter)
        stroke_width = 3

        angle1 = 0.58 * pi
        angle2 = 2.42 * pi
        radius = self._width / 2
        x0 = self.x + radius
        y0 = self.y + radius
        painter.draw_arc(x0, y0, radius, angle1, angle2,
                         width=stroke_width,
                         outline=(0, 0, 0))

        x1 = x0 - (radius * sin(0.08 * pi))
        y1 = y0 + (radius * cos(0.08 * pi))
        painter.draw_line(x0, y0, x1, y1, width=stroke_width, color=self.fgcolor())